#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Small script to compare arrival times of hybrid tau-p/fmm with simple tau-p for standard earth model

import os
import argparse

import numpy as np
import matplotlib.pyplot as plt

from obspy.taup import TauPyModel

from pylot.tomography.fmtomo_tools.fmtomo_teleseismic_utils import organize_receivers, organize_sources, organize_event_names


def read_file(fnin):
    infile = open(fnin, 'r')
    return infile.readlines()


def compare_arrivals(arrivals_file, receivers_file, sources_file, input_source_file, model='ak135_diehl_v2', exclude_phases=[]):
    '''
    Reads FMTOMO arrivals.dat file together with corresponding receiver, source and input_source files to match
    each arrival to a source and receiver combination, calculate tau-p arrivals for a given earth model and
    calculate differences
    '''
    arrivals_taup = {}
    arrivals_tomo = {}
    receiver_ids = {}
    event_names = {}
    events = organize_event_names(input_source_file)
    model = TauPyModel(model)

    # organize sources and receivers in dictionaries containing their ID used by FMTOMO as dictionary key
    receivers_dict = organize_receivers(receivers_file)
    sources_dict = organize_sources(sources_file)

    # read arrivals file
    with open(arrivals_file, 'r') as infile_arrivals:
        arrivals = infile_arrivals.readlines()

    count = 0
    for src_number, source in sources_dict.items():
        if source['phase'] in exclude_phases:
            continue
        src_name = events[src_number - 1]
        if not src_name in event_names.keys():
            arrivals_taup[src_name] = []
            arrivals_tomo[src_name] = []
            receiver_ids[src_name] = []
            event_names[src_name] = count
            count += 1

    for line in arrivals:
        # read line by line from fmtomo_tools output file arrivals.dat
        rec_id, src_id, ray_id, refl, arrival_time, diff, head = line.split()
        arrival_time = float(arrival_time)
        rec_id = int(rec_id)
        src_id = int(src_id)
        ray_id = int(ray_id)

        # identify receiver and source using dictionary
        receiver = receivers_dict[rec_id]
        source = sources_dict[src_id]
        src_name = events[src_id - 1]
        phase = source['phase']
        if phase in exclude_phases: continue
        taup_arrival = model.get_travel_times_geo(source_depth_in_km=source['depth'],
                                                  source_latitude_in_deg=source['lat'],
                                                  source_longitude_in_deg=source['lon'],
                                                  receiver_latitude_in_deg=receiver['lat'],
                                                  receiver_longitude_in_deg=receiver['lon'],
                                                  phase_list=[phase])
        receiver_depth_in_km = 6371. - receiver['rad']
        if len(taup_arrival) == 1:
            taup_arrival_time = taup_arrival[0].time
        else:
            taup_arrival_time = np.nan
        arrivals_taup[src_name].append(taup_arrival_time)
        arrivals_tomo[src_name].append(arrival_time)
        receiver_ids[src_name].append(rec_id)

    #plt.plot([min(arrivals_taup),max(arrivals_taup)],[min(arrivals_taup), max(arrivals_taup)], 'k-')
    sorted_by_first_arrival = sorted([(src_name, min(arrivals)) for src_name, arrivals in arrivals_taup.items()],
                                     key=lambda x: x[1])

    # print some output for analysis
    for item in sorted_by_first_arrival:
        print(item)
    #[print(source) for source in sources_dict.items()]
    #[print(item) for item in enumerate(events)]

    current_fmtomo_folder_name = os.path.split(os.path.abspath(arrivals_file))[-2]
    fname_savefig = '{}'.format(current_fmtomo_folder_name)
    if exclude_phases:
        fname_savefig += '_e'
        for phase in exclude_phases:
            fname_savefig += '_{}'.format(phase)

    plot_differences(arrivals_taup, arrivals_tomo, sorted_by_first_arrival, fname_savefig)
    #for event_name in ['20160124_103037.a_P.ttf', '20160729_211833.a_Pdiff.ttf', '20160729_211833.a_P.ttf']:
    #    plot_event(arrivals_tomo, arrivals_taup, receiver_ids, receivers_dict, src_name=event_name)


def plot_differences(arrivals_taup, arrivals_tomo, sorted_by_first_arrival, fname_savefig):
    fig = plt.figure(figsize=(16,9))
    ax = fig.add_subplot(111)

    # init plot for tt differences
    cmap = plt.get_cmap('jet')
    colors = cmap(np.linspace(0, 1, len(sorted_by_first_arrival)))

    for index, item in enumerate(sorted_by_first_arrival):
        src_name = item[0]
        ax.scatter(arrivals_taup[src_name], np.array(arrivals_tomo[src_name]) - np.array(arrivals_taup[src_name]),
                    c=colors[index], s=25, marker='.', label=src_name, edgecolors='none')

    # shrink box for legend
    box = ax.get_position()
    ax.set_position([box.x0, box.y0, box.width * 0.8, box.height])

    ax.legend(bbox_to_anchor=[1, 1], loc='upper left')

    plt.title(fname_savefig)
    ax.set_xlabel('Absolute time $t_{tau-p}$')
    ax.set_ylabel('Time difference $t_{hybrid} - t_{tau-p}$')

    print('Saving plot to {}.png'.format(fname_savefig))
    fig.savefig(fname_savefig + '.png', dpi=300)
    #plt.show()


def plot_event(arrivals_tomo, arrivals_taup, receiver_ids_dict, receivers_dict, src_name):
    arrivals_diff = np.array(arrivals_tomo[src_name]) - np.array(arrivals_taup[src_name])
    receiver_ids = receiver_ids_dict[src_name]
    x = np.array([receivers_dict[rec_id]['lon'] for rec_id in receiver_ids])
    y = np.array([receivers_dict[rec_id]['lat'] for rec_id in receiver_ids])
    sc = plt.scatter(x, y, c=arrivals_diff, edgecolor='none')
    plt.xlabel('Longitude [deg]')
    plt.ylabel('Latitude [deg]')
    cbar = plt.colorbar(sc)
    cbar.ax.set_ylabel('traveltime difference ($t_{hybrid} - t_{tau-p}$)')
    plt.title('{}'.format(src_name))
    plt.show()

# folders=[
# 'alparray_0_receiver_elev_zero',
# 'alparray_0_receiver_elev_zero_finer_pgrid_vgrid_r',
# 'alparray_0_receiver_elev_zero_finer_vgrid_llr',
# 'alparray_0_receiver_elev_zero_smaller_box',
# 'alparray_0_receiver_elev_zero_shallow_box',
# 'alparray_0_receiver_elev_zero_finer_pgrid_llr',
# 'alparray_0_receiver_elev_zero_finer_interface',
# #'alparray_0_receiver_elev_zero_bigger_box',
#     ]


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Compare arrivals with TauP-times')
    parser.add_argument('fmtomodir', help='path containing fm3d output')
    parser.add_argument('-e', dest='exclude', default=[], help='exclude phases, comma separated, no spaces')
    args = parser.parse_args()

    fdir = args.fmtomodir

    arrivals_file = os.path.join(fdir, 'arrivals.dat')
    receivers_file = os.path.join(fdir, 'receivers.in')
    sources_file = os.path.join(fdir, 'sources.in')
    input_source_file = os.path.join(fdir, 'input_source_file_P.in')

    exclude_phases = args.exclude
    if exclude_phases:
        exclude_phases = exclude_phases.split(',')
    compare_arrivals(arrivals_file, receivers_file, sources_file, input_source_file=input_source_file, exclude_phases=exclude_phases)


