#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os
from os.path import join as pjoin
import glob
import numpy as np
import warnings
import multiprocessing
from datetime import datetime

import json

import matplotlib.pyplot as plt

from functools import reduce

from scipy.interpolate import griddata, RegularGridInterpolator, interp1d
from scipy import signal

from obspy.taup.tau_model import TauModel

try:
    from pyvtk.hl import gridToVTK
except Exception as e:
    warnings.warn('Could not load pyevtk: {}'.format(e))

from pylot.tomography.utils import pol2cart, cart2pol
from pylot.tomography.fmtomo_tools.fmtomo_teleseismic_utils import Propgrid


def scale_deg(input_in_degree):
    scale_factor_deg_km = 2 * np.pi * 6371. / 360.
    return input_in_degree * scale_factor_deg_km

def propgrid2vtk(infile, outfile, R=6371., arrtimes=False):
    '''
    read and write a vtk file from FMTOMO propgrid.in file
    '''
    fid_in = open(infile, 'r')
    nR, nTheta, nPhi = [int(item) for item in fid_in.readline().split()]
    deltaR, deltaTheta_deg, deltaPhi_deg = [float(item) for item in fid_in.readline().split()]
    rtop, thetaS_deg, phiW_deg = [float(item) for item in fid_in.readline().split()]

    #deltaTheta = scale_deg(float(deltaTheta_deg))
    #deltaPhi = scale_deg(float(deltaPhi_deg))
    #thetaS = scale_deg(float(thetaS_deg))
    #phiW = scale_deg(float(phiW_deg))

    nPoints = nR * nTheta * nPhi

    out_id = open(outfile, 'w')
    out_id.write('# vtk DataFile Version 4.0\n')
    out_id.write('Propgrid\n')
    out_id.write('ASCII\n')
    #out_id.write('DATASET STRUCTURED_POINTS\n')
    #out_id.write('DIMENSIONS {nx:>15}{ny:>15}{nz:>15}\n'.format(nx=nTheta, ny=nPhi, nz=nR))
    #out_id.write('ORIGIN {x:>15}{y:>15}{z:>15}\n'.format(x=thetaS_deg, y=phiW_deg, z=rtop + R))
    #out_id.write('SPACING {dx:>15}{dy:>15}{dz:>15}\n'.format(dx=deltaTheta_deg, dy=deltaPhi_deg, dz=-deltaR))

    out_id.write('DATASET STRUCTURED_GRID\n')
    out_id.write('DIMENSIONS %15d %15d %15d\n' % (nR, nTheta, nPhi))
    out_id.write('POINTS %15d float\n' % (nPoints))

    lats = np.arange(thetaS_deg, stop=thetaS_deg + (nTheta) * deltaTheta_deg, step=deltaTheta_deg)
    lons = np.arange(phiW_deg, stop=phiW_deg + (nPhi) * deltaPhi_deg, step=deltaPhi_deg)
    rads = np.arange(rtop, stop=rtop - (nR) * deltaR, step=-deltaR)

    for lon in lons:
        for lat in lats:
            for r in rads:
                x, y, z = pol2cart(lat, lon, r)
                out_id.write('{x} {y} {z}\n'.format(x=x, y=y, z=z))

    fid_in.readline()
    fid_in.readline()

    if arrtimes:

        # write header with number of data points
        out_id.write('\nPOINT_DATA {}\n'.format(nPoints))

        # write header for data
        out_id.write('SCALARS {} float 1\n'.format('arrival_times'))
        out_id.write('LOOKUP_TABLE default\n')

        for _ in range(nPoints):
            out_id.write(fid_in.readline())
    #
    #
    # nPoints = nR * nTheta * nPhi
    #
    # out_id.write('VERTICES %15d %15d\n' % (nPoints, 2 * nPoints))
    #
    # # write indices
    # print("Writing indices to VTK file...")
    # for index in range(nPoints):
    #     out_id.write('%10d %10d\n' % (1, index))

    fid_in.close()
    out_id.close()
    print('Wrote file {}'.format(outfile))

def read_interfaces_file(infile):
    fid_in = open(infile, 'r')
    nInterfaces = int(fid_in.readline())
    nTheta, nPhi = [int(item) for item in fid_in.readline().split()] # INCLUDING CUSHION NODES (+2 each)
    deltaTheta_rad, deltaPhi_rad = fid_in.readline().split()
    thetaS_rad, phiW_rad = fid_in.readline().split() # INCLUDING CUSHION NODES (-delta each)

    deltaTheta = scale_deg(np.rad2deg(float(deltaTheta_rad)))
    deltaPhi = scale_deg(np.rad2deg(float(deltaPhi_rad)))
    thetaS = scale_deg(np.rad2deg(float(thetaS_rad)))
    phiW = scale_deg(np.rad2deg(float(phiW_rad)))

    interfaces = []
    for interface_index in range(nInterfaces):
        glob_index = 0
        interface = {'thetaGrid': np.empty(nTheta * nPhi),
                     'phiGrid': np.empty(nTheta * nPhi),
                     'points': np.empty(nTheta * nPhi)}
        for theta_index in range(nTheta):
            theta = thetaS + theta_index * deltaTheta
            for phi_index in range(nPhi):
                phi = phiW + phi_index * deltaPhi
                line = fid_in.readline().split()
                while len(line) == 0:
                    # skip blank lines
                    line = fid_in.readline().split()
                interface['thetaGrid'][glob_index] = theta
                interface['phiGrid'][glob_index] = phi
                interface['points'][glob_index] = float(line[0])
                glob_index += 1

        interfaces.append(interface)
    return interfaces

def interface2VTK(interface, filename='interface.vtk'):
    '''
    Generates a vtk file from all points of an interface
    '''
    outfile = open(filename, 'w')

    thetaGrid = interface['thetaGrid']
    phiGrid = interface['phiGrid']
    points = interface['points']
    ziped_grid = zip(thetaGrid, phiGrid, points)

    nPoints = len(ziped_grid)

    # write header
    print("Writing header for VTK file...")
    outfile.write('# vtk DataFile Version 3.1\n')
    outfile.write('Surface Points\n')
    outfile.write('ASCII\n')
    outfile.write('DATASET POLYDATA\n')
    outfile.write('POINTS %15d float\n' % (nPoints))

    # write coordinates
    print("Writing coordinates to VTK file...")

    for theta, phi, point in ziped_grid:
        outfile.write('%10f %10f %10f \n' % (theta, phi, point))

    outfile.write('VERTICES %15d %15d\n' % (nPoints, 2 * nPoints))

    # write indices
    print("Writing indices to VTK file...")
    for index in range(nPoints):
        outfile.write('%10d %10d\n' % (1, index))

    # outfile.write('POINT_DATA %15d\n' %(nPoints))
    # outfile.write('SCALARS traceIDs int %d\n' %(1))
    # outfile.write('LOOKUP_TABLE default\n')

    # # write traceIDs
    # print("Writing traceIDs to VTK file...")
    # for traceID in traceIDs:
    #     outfile.write('%10d\n' %traceID)

    outfile.close()
    print("Wrote %d points to file: %s" % (nPoints, filename))
    return

def interfaces2vtk(interfaces_file, outpath='.'):
    interfaces = read_interfaces_file(interfaces_file)
    for index, interface in enumerate(interfaces):
        interface2VTK(interface, os.path.join(outpath, 'interface_{}.vtk'.format(index + 1)))


def grid_differences(fname_in, fname_out, model, wdir='', percentage=True):
    # read external grid file
    #ext_grid = read_diehl_model(external_grid_file)
    #save_json_file(ext_grid, os.path.join(wdir, 'diehl_2009_grid.json'))

    # load/save external grid to json file (speeds up operation after saving to json once)
    ext_grid = load_json_file(os.path.join(wdir, fname_in))
    diff_grid = calculate_differences_grid_json_file(ext_grid, model, percentage=percentage)
    save_json_file(diff_grid, os.path.join(wdir, fname_out))
    outfile_vtk = fname_out.split('.json')[0]
    outfile_vtk += '.vtk'
    write_vtk(diff_grid, os.path.join(wdir, outfile_vtk), write_data=['vps'])


def interpolate_on_vgrid(vgrid_file, external_grid_file, fname_out, fname_out_vtk=None, ncores=None,
                         default_cov=1., vel_perturbation_abs=False, vel_perturbation_perc=False):
    if not ncores:
        ncores = multiprocessing.cpu_count()

    if vel_perturbation_abs and vel_perturbation_perc:
        raise IOError('Cannot be both abs and percentage perturbation.')

    ext_grid = load_json_file(external_grid_file)

    # re-sort grid contents to fit on vgrids.in
    a = np.array([ext_grid['lons'], ext_grid['lats'], ext_grid['depths']])
    ind = np.lexsort((a[0, :], a[1, :], a[2, ::-1]))
    for key, value in ext_grid.items():
        if value:
            ext_grid[key] = np.array(value)[ind]


    # write vtk file for external grid
    #write_vtk(ext_grid, fname=os.path.join(wdir, 'crustal_model.vtk'))

    # read FMTOMO velocity/inversion grid file
    vgrid, npts, delta, start = read_vgrid(vgrid_file)

    minDepth = min(ext_grid['depths'])
    maxDepth = max(ext_grid['depths'])
    minLat = min(ext_grid['lats'])
    maxLat = max(ext_grid['lats'])
    minLon = min(ext_grid['lons'])
    maxLon = max(ext_grid['lons'])

    depths = np.array(vgrid['depths'])
    lats = np.array(vgrid['lats'])
    lons = np.array(vgrid['lons'])

    # get indices of grid within min/max depths/lats/lons
    indices_depth = np.where(np.logical_and(depths >= minDepth, depths <= maxDepth))
    indices_lat = np.where(np.logical_and(lats >= minLat, lats <= maxLat))
    indices_lon = np.where(np.logical_and(lons >= minLon, lons <= maxLon))

    # get those indices common to all 3 indices arrays
    indices = reduce(np.intersect1d, (indices_depth, indices_lat, indices_lon))

    #x_vgrid = x[indices]
    #y_vgrid = y[indices]
    #z_vgrid = z[indices]

    input_list = prepare_multiprocessing_input(vgrid, ext_grid, indices, ncores)

    print('Generating multiprocessing pool with {} cores'.format(ncores))
    pool = multiprocessing.Pool(ncores)
    # interpolate on grid points
    interp_results = pool.map(griddata_worker, input_list)
    pool.close()
    print('Finished interpolation.')

    # calculate new vps_ext from differences
    #for index in indices:
    #    grid['vps'][index] += vps_new[index]

    vgrid['vps'] = np.array(vgrid['vps'])
    vgrid['covs'] = list(np.ones(len(vgrid['vps'])) * default_cov)
    for indices, new_vel in interp_results:
        if vel_perturbation_abs:
            vgrid['vps'][indices] += new_vel
        elif vel_perturbation_perc:
            vgrid['vps'][indices] *= 1 + new_vel/100.
        else:
            vgrid['vps'][indices] = new_vel
    vgrid['vps'] = list(vgrid['vps'])

    write_vgrid(vgrid, npts, delta, start, fname=fname_out)
    if fname_out_vtk:
        write_vtk(vgrid, fname_out_vtk, write_data=['vps', 'covs'], legacy=True)


def slice_array(array, indices, nparts):
    nparts = int(nparts)
    output_list = []
    nindices = len(indices)
    npercore = nindices//nparts
    if not nindices%nparts:
        npercore += 1
    for part in range(nparts):
        start_ind = part * npercore
        # slice indices array into parts for each core
        sub_indices = np.arange(start_ind, min([start_ind + npercore, nindices]), 1)
        output_list.append({'array': array[indices[sub_indices]], 'indices': indices[sub_indices]})
    return output_list


def prepare_multiprocessing_input(vgrid, ext_grid, indices, ncores):
    input_list = []
    input_dict = {}
    keys = ['xs', 'ys', 'zs']

    # slice all arrays into parts and store them in a dictionary
    for key in keys:
        array_list = slice_array(np.array(vgrid[key]), indices, nparts=ncores)
        input_dict[key] = array_list

    # now construct a list (one item for each core) with all 3 arrays in a dict and corresponding indices
    for index in range(ncores):
        d = {key: input_dict[key][index]['array'] for key in keys}
        # fetch indices from last used key (indices are all the same)
        d['indices'] = input_dict[key][index]['indices']
        # store input from external grid in dictionary (PROBLEM: might be very inefficient/memory consuming!)
        d['vps_ext'] = ext_grid['vps']
        d['x_ext'] = ext_grid['xs']
        d['y_ext'] = ext_grid['ys']
        d['z_ext'] = ext_grid['zs']
        input_list.append(d)

    return input_list


def griddata_worker(input_dict):
    data_velocity = np.array(input_dict['vps_ext'])
    data_coords = (np.array(input_dict['x_ext']), np.array(input_dict['y_ext']), np.array(input_dict['z_ext']))
    interp_coords = (input_dict['xs'], input_dict['ys'], input_dict['zs'])
    vps_new = griddata(data_coords, data_velocity, interp_coords, fill_value=0., method='nearest')
    return input_dict['indices'], vps_new


def calculate_differences_grid(grid, earth_model, veltype='p', percentage=True, input_rel=False):
    '''
    Calculate differences of input grid to standard earth model
    :param grid: dict
    :param earth_model: str
    :return: new grid with velocity differences
    '''

    abs_or_percentage = {True: 'percentage',
                         False: 'absolute'}
    print('Calculating {} differences to {} model for velocity type {}.'.format(abs_or_percentage[percentage],
                                                                                earth_model, veltype))

    veltype = veltype.lower()
    vel_key = 'v{}s'.format(veltype)

    model = TauModel.from_file(earth_model)

    for index, depth in enumerate(grid['depths']):
        if depth < 0.:
            depth = 0.
        try:
            vel_ref = model.s_mod.v_mod.evaluate_below(depth, veltype)
        except Exception as e:
            print('Exception at depth:', depth)
            raise(e)
        vel = grid[vel_key][index]
        if input_rel:
            vel += vel_ref
        if percentage:
            vel = (vel - vel_ref[0]) / vel_ref[0] * 100.
        else:
            vel -= vel_ref[0]
        grid[vel_key][index] = vel

    print('Done!')
    return grid


def calculate_differences_grid_json_file(grid, fn_earth_model, veltype='p', percentage=True):
    '''
    Calculate differences of input grid to standard earth model in json file
    :param grid: dict
    :param fn_earth_model: str
    :return: new grid with velocity differences
    '''

    abs_or_percentage = {True: 'percentage',
                         False: 'absolute'}
    print('Calculating {} differences to {} model for velocity type {}.'.format(abs_or_percentage[percentage],
                                                                                fn_earth_model, veltype))

    veltype = veltype.lower()
    vel_key = 'v{}s'.format(veltype)

    with open(fn_earth_model, 'r') as infile:
        model = json.load(infile)

    interp = interp1d(model['depths'], model['vps'])

    for index, depth in enumerate(grid['depths']):
        try:
            vel_ref = interp(depth)
        except Exception as e:
            print('Exception at depth:', depth)
            raise(e)
        vel = grid[vel_key][index]
        if percentage:
            vel = (vel - vel_ref) / vel_ref * 100.
        else:
            vel -= vel_ref
        grid[vel_key][index] = vel

    print('Done!')
    return grid


def convolution_filter(grid, sigma_xyz, spacing_xyz, R=6371., wdir='', fname_ext='', values_list=['vps']):
    '''
    Apply 3d convolution filter on grid
    :param grid: dict
    :param sigma_xyz: (s_x, s_y, s_z) std_dev in 3 dimensions for filter kernel
    :param R: earth radius
    :param wdir: working directory
    :return:
    '''
    if type(values_list) not in [list, tuple]:
        values_list = [values_list]

    npa = np.array

    lons_km = np.deg2rad(np.unique(grid['lons'])) * R * np.cos(np.deg2rad(46)) # reduce radius with mean at 46deg N (WHY???)
    lats_km = np.deg2rad(np.unique(grid['lats'])) * R
    zs = R - np.unique(grid['depths'])

    # re-sort grid contents
    a = np.array([grid['lons'], grid['lats'], grid['depths']])
    ind = np.lexsort((a[2, :], a[1, :], a[0, :]))
    for key, value in grid.items():
        if len(value) > 0:
            grid[key] = npa(value)[ind]
    # grid['lons'] = npa(grid['lons'])[ind]
    # grid['lats'] = npa(grid['lats'])[ind]
    # grid['depths'] = npa(grid['depths'])[ind]

    if len(grid['xs']) == 0 and len(grid['xs']) == 0 and len(grid['zs']) == 0:
        grid = calc_cartesian(grid)

    for val in values_list:
        if val == 'covs':
            print('Taking squareroot of covariances to convolve std, then square again.')
            values = np.sqrt(grid[val])
        else:
            values = grid[val]

        ## use indices to sort actual values and shape correspondingly (already done above!)
        #values = npa(values)[ind]
        shape = (len(lons_km), len(lats_km), len(zs))
        data_vel = values.reshape(shape)

        #data_vel = np.empty((len(lons_km), len(lats_km), len(zs)))
        #index = 0
        #for zind in range(len(zs)):
        #    for yind in range(len(lats_km)):
        #        for xind in range(len(lons_km)):
        #            data_vel[xind, yind, zind] = grid[val][index]
        #            index += 1

        #write_vtk(grid, os.path.join(wdir, 'grid_b4_filter{}_{}.vtk'.format(fname_ext, val)),
        #          write_data=[val], sort=True)

        gk = gauss_kernel(sigma_xyz, spacing_xyz=spacing_xyz)
        convolved_data_values = (signal.convolve(data_vel, gk, mode="same", method='fft') / np.sum(gk))

        val_new = convolved_data_values.ravel()

        if val == 'covs':
            print('Now square result again...')
            val_new = val_new ** 2

        grid[val] = val_new
        #index = 0
        #for zind in range(len(zs)):
        #    for yind in range(len(lats_km)):
        #        for xind in range(len(lons_km)):
        #            grid[val][index] = convolved_data_values[xind, yind, zind]
        #            index += 1

        #write_vtk(grid, os.path.join(wdir, 'grid_after_filter{}_{}.vtk'.format(fname_ext, val)),
        #          write_data=[val], sort=True)
    return grid


def calc_cartesian(grid, R=6371.):
    print('Calc cartesian coordinates for paraview...')
    lons = np.array(grid['lons'])
    lats = np.array(grid['lats'])
    depths = np.array(grid['depths'])
    xs, ys, zs = pol2cart(lats, lons, R - depths)
    grid['xs'] = xs
    grid['ys'] = ys
    grid['zs'] = zs
    return grid


# def convolution_filter_npy(data, sigma_llr, spacing_llr, R=6371., wdir='', fname_ext='', val='vps'):
#     '''
#     Apply 3d convolution filter on grid
#     :param grid: dict
#     :param sigma_llr: (s_x, s_y, s_z) std_dev in 3 dimensions for filter kernel
#     :param R: earth radius
#     :param wdir: working directory
#     :return:
#     '''
#     #lons_km = np.deg2rad(np.unique(grid['lons'])) * R
#     #lats_km = np.deg2rad(np.unique(grid['lats'])) * R # TODO This is shit?
#     #zs = R - np.unique(grid['depths'])
#
#     # data_vel = np.empty((len(lons_km), len(lats_km), len(zs)))
#     # index = 0
#     # for zind in range(len(zs)):
#     #     for yind in range(len(lats_km)):
#     #         for xind in range(len(lons_km)):
#     #             data_vel[xind, yind, zind] = grid[val][index]
#     #             index += 1
#
#     write_vtk(grid, os.path.join(wdir, 'grid_b4_filter{}.vtk'.format(fname_ext)), write_data=[val])
#
#     gk = gauss_kernel(sigma_llr, spacing_xyz=spacing_llr)
#     convolved_data_values = (signal.convolve(data_vel, gk, mode="same", method='fft') / np.sum(gk))
#
#     index = 0
#     for zind in range(len(zs)):
#         for yind in range(len(lats_km)):
#             for xind in range(len(lons_km)):
#                 grid[val][index] = convolved_data_values[xind, yind, zind]
#                 index += 1
#
#     write_vtk(grid, os.path.join(wdir, 'grid_after_filter{}.vtk'.format(fname_ext)), write_data=[val])
#     return grid


def gauss_kernel(sigma_xyz, spacing_xyz, width_factor_sigma=4):
    '''
    Calculate a three dimensional gauss kernel with a width of "width_factor_sigma" * sigma
    :param sigma_xyz: (sigma_x, sigma_y, sigma_z) standard deviation in x, y and z direction of kernel
    :param spacing_xyz: (spacing_x, spacing_y, spacing_z) spacing of grid to be convolved (important because
    convolution has no units!), e.g. Diehl grid: 2x2x2km spacing
    :param width_factor_sigma: size of the kernel (width_factor_sigma * sigma)
    :return: gauss kernel on a regular grid
    '''
    def calc_base_array(sigma, spacing):
        right_side = np.arange(0, width_factor_sigma * sigma, spacing)
        left_side = -1 * np.flip(right_side, 0)[:-1]
        return np.append(left_side, right_side)

    sigma_x, sigma_y, sigma_z = sigma_xyz
    spacing_x, spacing_y, spacing_z = spacing_xyz
    #assert type(npts) == int, 'Wrong input type {} for npts'.format(type(npts))
    #if npts%2:
    #    npts += 1
    #    warnings.warn('Npts has to be odd for symmetric kernel. Increased by + 1')
    x = np.copy(calc_base_array(sigma_x, spacing_x))
    y = np.copy(calc_base_array(sigma_y, spacing_y))
    z = np.copy(calc_base_array(sigma_z, spacing_z))
    xx, yy, zz = np.meshgrid(x, y, z)
    #norm_factor = 1. / (2. * np.sqrt(2.) * np.pi ** (3. / 2.) * sigma ** 3)
    #gauss = norm_factor * np.exp(-(xx ** 2 + yy ** 2 + zz ** 2) / (2. * sigma ** 2))
    gauss = np.exp((-1. / 2.) * (xx ** 2 / (sigma_x ** 2) + yy ** 2 / (sigma_y ** 2) + zz ** 2 / sigma_z ** 2))
    return gauss


def save_json_file(grid, fname):
    for key, value in grid.items():
        if not type(value) == list:
            grid[key] = list(value)
    with open(fname, 'w') as outfile:
        json.dump(grid, outfile)
    print('Saved grid to file {}'.format(fname))


def load_json_file(fname):
    with open(fname, 'r') as infile:
        grid = json.load(infile)
    print('Loaded grid from file {}'.format(fname))
    return grid


def write_vgrid(vgrid, npts, delta, start, fname, default_cov=0.3, default_smooth=1.0, default_pdev=0.):
    '''
    Write vgrids input file for fmtomo_tools.
    :param vgrid: grid dictionary
    :param npts: number of grid points (3dim)
    :param delta: spacings (3dim)
    :param start: start points (3dim)
    :param fname: outfile
    :param cov: diagonal elements of covariance matrix
    :return:
    '''
    nR, nTheta, nPhi = npts
    dR, dTheta, dPhi = delta
    sR, sTheta, sPhi = start

    if not vgrid.get('covs'):
        print('No covariances given. Setting every point to default covariance of {}'.format(default_cov))
        vgrid['covs'] = list(default_cov * np.ones(len(vgrid['vps'])))

    if not vgrid.get('smoothfactors'):
        print('No smoothfactors given. Setting every point to default smoothfactor of {}'.format(default_smooth))
        vgrid['smoothfactors'] = list(default_smooth * np.ones(len(vgrid['vps'])))

    if not vgrid.get('frechs'):
        print('No pdev given. Setting every point to default pdev of {}'.format(default_pdev))
        vgrid['frechs'] = list(default_pdev * np.ones(len(vgrid['vps'])))

    with open(fname, 'w') as outfile:
        # write header for velocity grid file (in RADIANS)
        outfile.write('%10s %10s \n' % (1, 1))
        outfile.write('%10s %10s %10s\n' % (nR, nTheta, nPhi))
        outfile.write('%10s %10s %10s\n' % (dR, np.deg2rad(dTheta), np.deg2rad(dPhi)))
        outfile.write('%10s %10s %10s\n' % (sR, np.deg2rad(sTheta), np.deg2rad(sPhi)))
        
        for vel, cov, smoothfactor, pdev in zip(vgrid['vps'], vgrid['covs'], vgrid['smoothfactors'], vgrid['frechs']):
            outfile.write('{}    {}    {}    {}\n'.format(vel, cov, smoothfactor, pdev))

    print('Wrote {} points to file {}'.format(nR * nTheta * nPhi, fname))


def read_vgrid_regular(grid_file, R=6371.):
    '''
    Read fmtomo_tools velocity grid file (vgrids.in) in a regular grid structure
    as used by scipy.interpolate.RegularGridInterpolator
    :param grid_file: filename of fmtomo_tools velocity grid file
    :param R: earth radius
    :return: grid, vps, covariances, smoothings, partial_derivs
    '''
    gridN, gridDelta, gridStart, velocities, covariances, smooths, pdevs = _readVgrid(grid_file)
    lats, lons, rads = _generateGrids(gridN, gridDelta, gridStart)

    grid = (lons, lats, rads)

    vps = np.zeros((len(lons), len(lats), len(rads)))
    covs = np.zeros((len(lons), len(lats), len(rads)))
    sms = np.zeros((len(lons), len(lats), len(rads)))
    pds = np.zeros((len(lons), len(lats), len(rads)))

    index = 0
    for irad, rad in enumerate(rads):
        for ilat, lat in enumerate(lats):
            for ilon, lon in enumerate(lons):
                vps[ilon, ilat, irad] = velocities[index]
                cov = np.nan if not covariances else covariances[index]
                smooth = np.nan if not smooths else smooths[index]
                pdev = np.nan if not pdevs else pdevs[index]

                covs[ilon, ilat, irad] = cov
                sms[ilon, ilat, irad] = smooth
                pds[ilon, ilat, irad] = pdev
                index += 1

    return grid, vps, covs, sms, pds


def read_vgrid(grid_file, R=6371., inv_index_frechet=False):
    '''
    Read fmtomo_tools velocity grid file (vgrids.in) to "grid" dictionary structure.
    :param grid_file: filename of fmtomo_tools velocity grid file
    :param R: earth radius
    :return: grid, Npoints (3dim), delta (3dim), start (3dim)
    '''
    gridN, gridDelta, gridStart, velocities, covariances, smooths, pdevs = _readVgrid(grid_file)
    lats, lons, rads = _generateGrids(gridN, gridDelta, gridStart)

    grid = init_dict()

    if inv_index_frechet:
        nR, nLat, nLon = gridN
        grid['inv_index'] = list(np.zeros(nR * nLat * nLon))

    covariance_flag = True if len(covariances) == len(velocities) else False
    if covariance_flag:
        print('Found covariances in velocity model')
    else:
        print('No covariances found in velocity model.')


    index = 0
    for r in rads:
        for lat in lats:
            for lon in lons:
                # get values from grid file and append to grid dictionary
                grid['lons'].append(lon)
                grid['lats'].append(lat)
                grid['depths'].append(R - r)

                # transform to cartesian coordinate system and append to grid dictionary
                x, y, z = pol2cart(lat, lon, r)

                grid['xs'].append(x)
                grid['ys'].append(y)
                grid['zs'].append(z)

                grid['vps'].append(velocities[index])

                if covariance_flag:
                    grid['covs'].append(covariances[index])

                # NOTE: inv_index only valid in case of inversion using one velocity grid, else there would be an offset for each
                # other parameter/inversion grid
                # inversion index (starts with 1) is built differently: for lon in... for lat in... for r in...
                if inv_index_frechet:
                    glob_inv_index = transform_point_index(index, nR, nLat, nLon)
                    grid['inv_index'][glob_inv_index] = index

                index += 1

    return grid, gridN, gridDelta, gridStart


def add_cushion_zeros_top_bot(fname_in, fname_out, nlayers, R=6371., top=False, bot=True):
    def depth_layer(lons, lats, depth):
        subgrid = dict(depths=[], lats=[], lons=[], xs=[], ys=[], zs=[], vps=[])
        # note: diehl model counts all longitudes first (negative direction), then latitude (positive direction),
        # then depth (positive direction)
        for lat in lats:
            for lon in reversed(lons):
                subgrid['depths'].append(depth)
                subgrid['lats'].append(lat)
                subgrid['lons'].append(lon)

                # transform to cartesian coordinate system and append to grid dictionary
                x, y, z = pol2cart(lat, lon, R - depth)
                subgrid['xs'].append(x)
                subgrid['ys'].append(y)
                subgrid['zs'].append(z)

                # append zeros
                subgrid['vps'].append(0.)
        return subgrid

    assert any([top, bot]), 'No sense running this function...'

    with open(fname_in, 'r') as infile:
        grid = json.load(infile)

    print('Read {} points from file {}'.format(len(grid['depths']), fname_in))

    lons = np.unique(grid['lons'])
    lats = np.unique(grid['lats'])


    # add cushion n-times
    for index in range(nlayers):
        # calculate new depths for iterating process
        depths = np.unique(grid['depths'])
        if index == 0:
            ddepth = abs(depths[1] - depths[0])

        if top == True:
            top_depth = min(depths) - ddepth
            top_layer = depth_layer(lons, lats, top_depth)
            for key in top_layer.keys():
                # add top layer to beginning of list
                grid[key] = top_layer[key] + grid[key]

        if bot == True:
            bot_depth = max(depths) + ddepth
            bot_layer = depth_layer(lons, lats, bot_depth)
            for key in bot_layer.keys():
                # add bot layer to end of list
                grid[key] = grid[key] + bot_layer[key]

    with open(fname_out, 'w') as outfile:
        json.dump(grid, outfile)
    print('Wrote file', fname_out)


def read_diehl_model(grid_file, R=6371.):
    with open(grid_file, 'r') as infile:
        input_list = infile.readlines()

    grid = init_dict()

    for line in input_list:
        # get values from grid file and append to grid dictionary
        lon, lat, x_simulp, y_simulp, depth, vel = [float(item) for item in line.split()]
        grid['lons'].append(lon)
        grid['lats'].append(lat)
        grid['depths'].append(depth)
        grid['vps'].append(vel)

        # transform to cartesian coordinate system and append to grid dictionary
        x, y, z = pol2cart(lat, lon, R - depth)
        grid['xs'].append(x)
        grid['ys'].append(y)
        grid['zs'].append(z)

    return grid


def read_tesauro_model(grid_file, min_depth, max_depth, npts_depth, R=6371.):
    with open(grid_file, 'r') as infile:
        input_list = infile.readlines()[1:]

    grid = init_dict()

    depths = np.linspace(min_depth, max_depth, num=npts_depth)

    for depth in depths:
        for line in input_list:
            # get values from grid file and append to grid dictionary
            lon, lat, uc_vel, lc_vel, avcrust, topo, base, uc_lc, moho = [float(item) for item in line.split()]
            grid['lons'].append(lon)
            grid['lats'].append(lat)
            grid['depths'].append(depth)
            if depth <= topo:
                vel = uc_vel
                cov = 0.05
            elif topo < depth <= uc_lc:
                vel = uc_vel
                cov = 0.1
            elif uc_lc < depth <= moho:
                vel = lc_vel
                cov = 0.1
            else:
                vel = 8.0355 # TODO: BETTER USE NAN/ak135 vel?
                cov = 1.

            grid['vps'].append(vel)
            grid['covs'].append(cov)

            # transform to cartesian coordinate system and append to grid dictionary
            x, y, z = pol2cart(lat, lon, R - depth)
            grid['xs'].append(x)
            grid['ys'].append(y)
            grid['zs'].append(z)

    return grid


def add_array_layer(a, depth_val, depth_ind):
    inds = np.lexsort((a[0, :], a[1, :], a[depth_ind, :]))
    a = a[:, inds]
    inds_tl = np.where(a[depth_ind, :] == np.unique(a[depth_ind, :])[0])[0]
    if not all(np.diff(inds_tl) == 1):
        print('Adding layer failed.')
        return a
    end_ind = len(inds_tl)
    cp_layer = np.copy(a[:, :end_ind])
    cp_layer[depth_ind, :] = depth_val
    return np.append(cp_layer, a, axis=1)


def get_rginter_tesauro_discrete(fnin, fill_value=np.nan, relative_values=False, ftype='npy', col_index=3,
                                 extend_surface=15, depth_ind=2, min_depth=None, max_depth=None):
    '''
    Create and return rginter object containing vp data of discrete Tesauro grid.
    :param fnin: fname tesauro discrete (txt)
    :param extend_surface: Extend uppermost value of external grid to x km above sea level
    :return: rginter object accessed by: rginter((lon, lat, depth))
    '''
    if ftype == 'txt':
        a = np.loadtxt(fnin).transpose()
    elif ftype == 'npy':
        a = np.load(fnin)

    if extend_surface:
        a = add_array_layer(a, -extend_surface, depth_ind)

    # sort by depth, lat, lon (outer loop = lon)
    inds = np.lexsort((a[2, :], a[1, :], a[0, :]))
    a = a[:, inds]

    lons, lats, depths = a[:3]
    values = a[col_index]

    if not relative_values:
        # negative velocities -> np.nan
        values[values < 0] = np.nan

    lonsU = np.unique(lons)
    latsU = np.unique(lats)
    depthsU = np.unique(depths)
    grid = (lonsU, latsU, depthsU) #TODO shapes kontrollieren
    shape = [len(item) for item in grid]

    if min_depth:
        inds_lower_mindepth = np.where(depths < min_depth)[0]
        values[inds_lower_mindepth] = np.nan

    if max_depth:
        inds_over_maxdepth = np.where(depths > max_depth)[0]
        values[inds_over_maxdepth] = np.nan

    values = values.reshape(shape)
    rginter = RegularGridInterpolator(grid, values, method='linear', bounds_error=False, fill_value=fill_value)
    return rginter


def get_rginter_tesauro_discrete_std(fnin, std_in=0.3162, std_out=1.0, ftype='npy'):
    '''
    Create and return rginter object containing cov data based on vp of discrete Tesauro grid. If vp is negative (outside
    of grid) std will be rde_out, else rde_in.
    :param fnin: fname tesauro discrete (txt)
    :return: rginter object accessed by: rginter((lon, lat, depth))
    '''
    if ftype == 'txt':
        lons, lats, depths, values = np.loadtxt(fnin).transpose()
    elif ftype == 'npy':
        lons, lats, depths, values = np.load(fnin)
    # negative velocities -> np.nan
    values[values < 0] = std_out
    values[values >= 0] = std_in
    lonsU = np.unique(lons)
    latsU = np.unique(lats)
    depthsU = np.unique(depths)
    grid = (lonsU, latsU, depthsU)
    shape = [len(item) for item in grid]
    values = values.reshape(shape)
    rginter = RegularGridInterpolator(grid, values, method='linear', bounds_error=False, fill_value=std_out)
    return rginter


def interpolate_ray_crossing_on_vgrid(fname_crossing='vgrids_v5_it12_crossing_bins_4-1_mincount_5_STEP_2.npy',
                                      fname_vg='it_12/vgrids.in', fname_vgref='vgridsref_orig.in', R=6371.):
    '''
    Interpolate resolution information given in fname_crossing (.npy) onto velocity grid in fname_vg. Also calculate
    dvp (percentage). Return data as numpy array.
    '''
    cross_grid = np.load(fname_crossing)
    cross_grid[2, :] = R - cross_grid[2, :]
    inds_a = np.lexsort([cross_grid[0, :], cross_grid[1, :], cross_grid[2, :]])
    cross_grid = cross_grid[:, inds_a]
    lons, lats, depths, res = cross_grid

    lonsU = np.unique(lons)
    latsU = np.unique(lats)
    depthsU = np.unique(depths)
    grid = (depthsU, latsU, lonsU)
    shape = [len(item) for item in grid]
    values = res.reshape(shape)
    rginter = RegularGridInterpolator(grid, values, method='linear', bounds_error=False, fill_value=0)

    vg = read_vgrid(fname_vg)[0]
    vg = np.array([(vg['lons'], vg['lats'], vg['depths'], vg['vps'])])[0]
    inds_b = np.lexsort([vg[0, :], vg[1, :], vg[2, :]])
    vg = vg[:, inds_b]
    lonsVG, latsVG, depthsVG, vpsVG = vg

    vgref = read_vgrid(fname_vgref)[0]
    vgref = np.array([(vgref['lons'], vgref['lats'], vgref['depths'], vgref['vps'])])[0]
    vgref = vgref[:, inds_b]
    vpsVGREF = vgref[3]
    dvp = (vpsVG - vpsVGREF) / vpsVGREF * 100.
    vg = np.vstack([vg, dvp])

    resVG = rginter((depthsVG, latsVG, lonsVG))
    vg = np.vstack([vg, resVG])

    # remove velocity information where no resolution is given:
    #inds_no_res = np.where(resVG == 0.)[0]
    #vg[3, inds_no_res] = vgref[3, inds_no_res]
    #vg[4, inds_no_res] = 0.

    return vg


# def get_rginter_tesauro_discrete_rde(fnin, rde_in=0.1, rde_out=0., ftype='npy'):
#     '''
#     Create and return rginter object containing cov data based on vp of discrete Tesauro grid. If vp is negative (outside
#     of grid) std will be rde_out, else rde_in.
#     :param fnin: fname tesauro discrete (txt)
#     :return: rginter object accessed by: rginter((lon, lat, depth))
#     '''
#     if ftype == 'txt':
#         lons, lats, depths, values = np.loadtxt(fnin).transpose()
#     elif ftype == 'npy':
#         lons, lats, depths, values = np.load(fnin)
#     # negative velocities -> np.nan
#     #WRONG, vp +/-
#     #values[values < 0] = rde_out
#     #values[values >= 0] = rde_in
#     lonsU = np.unique(lons)
#     latsU = np.unique(lats)
#     depthsU = np.unique(depths)
#     grid = (lonsU, latsU, depthsU)
#     shape = [len(item) for item in grid]
#     values = values.reshape(shape)
#     rginter = RegularGridInterpolator(grid, values, method='linear', bounds_error=False, fill_value=rde_out)
#     return rginter


def get_index(point_index, nj, nk):
    '''
    Small function to calculate i,j,k indices for a point_index counting over i, then j, then k
    '''
    i = point_index // (nj * nk)
    if point_index % (nj * nk):
        i += 1
        j = point_index % (nj * nk) // nk
    else:
        j = nj
    if point_index % (nj * nk) % nk:
        j += 1
        k = point_index % (nj * nk) % nk
    else:
        k = nk
    return i, j, k


def transform_point_index(point_index, ni, nj, nk):
    '''
    Small function to calculate new point index counting over k, then j, then i instead of i, j, k
    '''
    i, j, k = get_index(point_index, nj, nk)
    new_point_index = (nj * ni) * (k - 1) + ni * (j - 1) + i
    return new_point_index


#def write_vtk_pyevtk(grid, fname, write_cov=False):
#    cellData = {'velocity': np.array(grid['vps'])}
#    if write_cov:
#        cellData['diag_covariance'] = np.array(grid['covs'])
#
#    gridToVTK(fname, np.array(grid['xs'], dtype='float64'), np.array(grid['ys'], dtype='float64'),
#              np.array(grid['zs'], dtype='float64'), cellData={'zs': np.array(grid['zs'], dtype='float64')})


def write_vtk_legacy(grid, fname, write_data=['vps']):
    with open(fname, 'w') as outfile:
        nPoints = len(grid['xs'])

        outfile.write('# vtk DataFile Version 3.1\n')
        outfile.write('FMM Init points\n')
        outfile.write('ASCII\n')
        nTheta = len(np.unique(grid['lats']))
        nPhi = len(np.unique(grid['lons']))
        nR = len(np.unique(grid['depths']))

        outfile.write('DATASET STRUCTURED_GRID\n')
        outfile.write('DIMENSIONS %15d %15d %15d\n' % (nPhi, nTheta, nR))
        outfile.write('POINTS %15d float\n' % (nPoints))

        for x, y, z in zip(grid['xs'], grid['ys'], grid['zs']):
            outfile.write('{x} {y} {z}\n'.format(x=x, y=y, z=z))

        # write header with number of data points
        outfile.write('\nPOINT_DATA {}\n'.format(nPoints))

        if 'grid_indices' in write_data:
            grid['grid_indices'] = list(range(len(grid['xs'])))

        for data_name in write_data:
            if not data_name in grid.keys():
                warnings.warn('Data with name: {} not found in grid.'.format(data_name))
                continue

            # translate data name to a label for convenience
            translate_names = {'vps': 'p_velocity',
                               'covs': 'diag_covariance',
                               'frechs': 'frechet_deriv_sum',
                               'rdes': 'res_mat_diagonal_element',
                               'res': 'resolution_information'}
            if data_name in translate_names.keys():
                data_label = translate_names[data_name]
            else:
                data_label = data_name

            # write header for data
            outfile.write('SCALARS {} float 1\n'.format(data_label))
            outfile.write('LOOKUP_TABLE default\n')

            for datum in grid[data_name]:
                if np.isnan(datum): datum = -1.
                outfile.write('{}\n'.format(datum))

    print('Wrote {} points to file {}'.format(nPoints, fname))


def write_vtk(grid, fname, write_data=['vps'], clat=None, clon=None, dlat=None, dlon=None, legacy=True, sort=False):
    if sort:
        import copy
        grid = copy.deepcopy(grid)
        grid = re_sort_grid(grid, write_data)

    if clat and clon and dlat and dlon:
        lat = np.array(grid['lats'])
        lon = np.array(grid['lons'])
        ids = np.where((lat > clat - dlat) & (lat < clat + dlat) & (lon > clon - dlon) & (lon < clon + dlon))
        for key, value in grid.items():
            if value:
                grid[key] = list(np.array(value)[ids])

    #if legacy:
    write_vtk_legacy(grid, fname, write_data)
    #else:
    #    write_vtk_pyevtk(grid, fname, write_cov)


def re_sort_grid(grid, write_data):
    grid_keys = ['lons', 'lats', 'depths', 'xs', 'ys', 'zs'] + [val for val in write_data]
    array = np.array([grid[key] for key in grid_keys])

    # rearrange array by different rows (outer loop inverse depth [2, ::-1] then lat [1, :] then lon [0, :]) for vtk
    ind = np.lexsort((array[0, :], array[1, :], array[2, ::-1]))
    array = array[:, ind]

    for index, key in enumerate(grid_keys):
        grid[key] = list(array[index, :])

    return grid


def _readVgrid(filename):
    def readNumberOfPoints(filename):
        fin = open(filename, 'r')
        vglines = fin.readlines()

        nR = int(vglines[1].split()[0])
        nTheta = int(vglines[1].split()[1])
        nPhi = int(vglines[1].split()[2])

        print('readNumberOf Points: Awaiting %d grid points in %s'
              % (nR * nTheta * nPhi, filename))
        fin.close()
        return nR, nTheta, nPhi

    def readDelta(filename):
        fin = open(filename, 'r')
        vglines = fin.readlines()

        dR = float(vglines[2].split()[0])
        dTheta = float(vglines[2].split()[1])
        dPhi = float(vglines[2].split()[2])

        fin.close()
        return dR, dTheta, dPhi

    def readStartpoints(filename):
        fin = open(filename, 'r')
        vglines = fin.readlines()

        sR = float(vglines[3].split()[0])
        sTheta = float(vglines[3].split()[1])
        sPhi = float(vglines[3].split()[2])

        fin.close()
        return sR, sTheta, sPhi

    def readVelocity(filename):
        '''
        Reads in velocity from vgrids file and returns a list containing all values in the same order
        '''
        vel = []
        cov = []
        smooths = []
        pdevs = []
        count = 0
        fin = open(filename, 'r')
        vglines = fin.readlines()

        for line in vglines:
            count += 1
            if count > 4:
                l_split = line.split()
                if l_split:
                    vel.append(float(l_split[0]))
                    if len(l_split) >= 2:
                        cov.append(float(l_split[1]))
                    if len(l_split) >= 3:
                        smooths.append(float(l_split[2]))
                    if len(l_split) >= 4:
                        pdevs.append(float(l_split[3]))


        print("Read %d points out of file: %s" % (len(vel), filename))
        return vel, cov, smooths, pdevs

    # Theta, Phi in radians, R in km
    nR, nTheta, nPhi = readNumberOfPoints(filename)
    dR, dThetaRad, dPhiRad = readDelta(filename)
    sR, sThetaRad, sPhiRad = readStartpoints(filename)
    vel, cov, smooths, pdevs = readVelocity(filename)

    dTheta, dPhi = np.rad2deg((dThetaRad, dPhiRad))
    sTheta, sPhi = np.rad2deg((sThetaRad, sPhiRad))

    number = (nR, nTheta, nPhi)
    delta = (dR, dTheta, dPhi)
    start = (sR, sTheta, sPhi)
    return number, delta, start, vel, cov, smooths, pdevs


def _generateGrids(number, delta, start):
    nR, nTheta, nPhi = number
    dR, dTheta, dPhi = delta
    sR, sTheta, sPhi = start

    eR = sR + (nR - 1) * dR
    ePhi = sPhi + (nPhi - 1) * dPhi
    eTheta = sTheta + (nTheta - 1) * dTheta

    thetaGrid = np.linspace(sTheta, eTheta, num=nTheta)
    phiGrid = np.linspace(sPhi, ePhi, num=nPhi)
    rGrid = np.linspace(sR, eR, num=nR)

    return (thetaGrid, phiGrid, rGrid)


def apply_convolution_filter(sigx, sigy, sigz, spacing_xyz, infile, wdir, fname_out, values=['vps'], io_json=False):
    if io_json:
        fname_ext = '_{}'.format(os.path.split(fname_out)[-1].split('.json')[0])
        grid = load_json_file(infile)
    else:
        list_in = np.load(infile)
        lons, lats, depths = list_in[:3]
        values_dict = {key: list_in[index + 3] for index, key in enumerate(values)}
        grid = grid_from_latlon_values(lons, lats, depths, values_dict=values_dict)
        fname_ext = ''
    grid = convolution_filter(grid, (sigx, sigy, sigz), spacing_xyz, wdir=wdir, fname_ext=fname_ext, values_list=values)
    if io_json:
        save_json_file(grid, fname_out)
    else:
        a = np.array([grid['lons'], grid['lats'], grid['depths'], *[grid[val] for val in values]])
        np.save(fname_out, a)


def apply_vel_changes_to_absolute_grid(fname_abs, fname_rel, fname_out, perc=True):
    grid_abs = load_json_file(fname_abs)
    grid_rel = load_json_file(fname_rel)
    # account for grid_rel being extended vertically
    depths = np.unique(grid_abs['depths'])
    vel_abs = np.array(grid_abs['vps'])
    vel_rel = np.array(grid_rel['vps'])[(np.array(grid_rel['depths']) >= min(depths)) & (np.array(grid_rel['depths']) <= max(depths))]
    if perc:
        vel_abs *= 1 + vel_rel / 100
    else:
        vel_abs += vel_rel
    grid_abs['vps'] = list(vel_abs)
    save_json_file(grid_abs, fname_out)


def init_dict():
    grid = dict(lons = [], lats = [], depths = [], vps = [], covs = [], xs = [], ys = [], zs = [])
    return grid


def points_from_vtk(fname_in, unique_points=True):
    points = []
    # assuming this value is always the same...
    points_per_line = 3
    with open(fname_in, 'r') as infile:
        header = infile.readline()
        print('Reading file {}: {}'.format(fname_in, header))
        header2 = infile.readline()
        dtype = infile.readline()
        _, dtype = infile.readline().split()
        assert(dtype == 'POLYDATA'), 'Expecting POLYDATA'
        pts, npts, pts_type = infile.readline().split()
        assert(pts == 'POINTS'), 'Expecting POINTS'
        for index in range(int(npts)//points_per_line):
            items = infile.readline().split()
            index_start = 0
            for lineindex in range(points_per_line):
                point = [float(item) for item in items[index_start: index_start + points_per_line]]
                if not point in points or not unique_points:
                    points.append(point)
                index_start += points_per_line

    return np.array(points)


def point_in_box(box_corner_points, points_to_check):
    def check_dependancy(u, v, w):
        if abs(np.linalg.det(np.array([u, v, w]))) > 1e-7:
            return False
        else:
            return True

    p0 = box_corner_points[0]
    p1 = box_corner_points[1]
    p2 = box_corner_points[2]
    p4 = box_corner_points[4]

    u = p1 - p0
    v = p2 - p0
    w = p4 - p0

    assert(check_dependancy(u, v, w) == False), 'Vectors u, v, w seem to be linear dependent: {}, {}, {}'.format(u, v, w)

    print('Checking {} points.'.format(len(points_to_check)))
    results = []
    for index, point in enumerate(points_to_check):
        if all([np.dot(u, p0) <= np.dot(u, point) <= np.dot(u, p1),
                np.dot(v, p0) <= np.dot(v, point) <= np.dot(v, p2),
                np.dot(w, p0) <= np.dot(w, point) <= np.dot(w, p4)]):
            results.append(True)
        else:
            results.append(False)

    return np.array(results)


def discretize_box(fname_in, fname_out, vp_diff=3., R=6371., dr=10., dLat=0.1, dLon=0.1, cf=5, earth_model='ak135_diehl_v2'):
    '''

    :param fname_in: vtk input filename generated from a box in Paraview
    :param fname_out: output parameter (.json!)
    :param vp_diff: p_velocity increase/decrease (%)
    :param R: earth radius
    :param dr: delta R
    :param dLat: delta Latitude
    :param dLon: delta Longitude
    :param cf: cushion factor (add some spacing outside the box boundaries for the Gauss Kernel)
    :return:
    '''
    points = points_from_vtk(fname_in)
    corner_grid = init_dict()
    box_grid = init_dict()

    for x, y, z in points:
        corner_grid['xs'].append(x)
        corner_grid['ys'].append(y)
        corner_grid['zs'].append(z)

        # transform from cartesian coordinate system and append to corner_grid dictionary
        lat, lon, r = cart2pol(x, y, z)
        corner_grid['lons'].append(lon)
        corner_grid['lats'].append(lat)
        corner_grid['depths'].append(R - r)

    print('Initialized corner grid')

    minDepth = min(corner_grid['depths'])
    maxDepth = max(corner_grid['depths'])
    minLat = min(corner_grid['lats'])
    maxLat = max(corner_grid['lats'])
    minLon = min(corner_grid['lons'])
    maxLon = max(corner_grid['lons'])

    depths = np.arange(minDepth - cf * dr, maxDepth + cf * dr, step=dr)
    lats = np.arange(minLat - cf * dLat, maxLat + cf * dLat, step=dLat)
    lons = np.arange(minLon - cf * dLon, maxLon + cf * dLon, step=dLon)

    print('NLat, NLon, Ndepth:', len(lats), len(lons), len(depths))

    points_to_check = []
    for depth in depths:
        for lat in lats:
            for lon in lons:
                box_grid['depths'].append(depth)
                box_grid['lats'].append(lat)
                box_grid['lons'].append(lon)

                # transform to cartesian coordinate system and append to box_grid dictionary
                x, y, z = pol2cart(lat, lon, R - depth)
                box_grid['xs'].append(x)
                box_grid['ys'].append(y)
                box_grid['zs'].append(z)

                points_to_check.append([x, y, z])

                # append zeros
                box_grid['vps'].append(0.)

    print('Initialized box grid')

    box_corner_points = np.array(list(zip(corner_grid['xs'], corner_grid['ys'], corner_grid['zs'])))
    pib = point_in_box(box_corner_points, np.array(points_to_check))

    print('Calculating final grid...')

    model = TauModel.from_file(earth_model)


    for index in range(len(box_grid['vps'])):
        if pib[index]:
            depth = box_grid['depths'][index]
            vel_ref = model.s_mod.v_mod.evaluate_below(depth, 'P')[0]
            box_grid['vps'][index] = vp_diff / 100. * vel_ref

    with open(fname_out, 'w') as outfile:
        json.dump(box_grid, outfile)
    print('Wrote file {}'.format(fname_out))


def read_diehl_rde_layers(inputdir='/data/AlpArray_Data/crust_diehl_2009/Diehl_et_al_2009_Alpine_HorizontalXSections/',
                          sigma_xyz=(12.5, 12.5, 12.5), spacing_xyz=(2, 2, 15)):
    depths = {infile: float(infile.replace('.xyz', '').lstrip('plane').lstrip('0')) for infile in
              sorted(glob.glob1(inputdir, '*.xyz'))}
    grid = init_dict()
    grid['rdes'] = []
    for infile, depth in depths.items():
        print('Working on file:', infile)
        points = np.loadtxt(os.path.join(inputdir, infile), skiprows=2)
        lons = points[:, 2]
        lats = points[:, 3]
        rdes = points[:, 11]
        grid['lons'] += list(lons)
        grid['lats'] += list(lats)
        grid['depths'] += list(np.ones(len(lons))*depth)
        grid['rdes'] += list(rdes)
    return grid
    #grid_conv = convolution_filter(grid, sigma_llr=sigma_llr, spacing_llr=spacing_llr, val='rdes')
    #return grid_conv


def get_rginter_diehl_legacy(grid, value='vps', max_value=np.inf, min_value=-np.inf, fill_value=0):
    print('Create rginter Diehl')
    lats = np.unique(grid['lats'])
    lons = np.unique(grid['lons'])
    depths = np.unique(grid['depths'])
    values = np.zeros((len(lons), len(lats), len(depths)))
    index = 0
    for idepth, depth in enumerate(depths):
        for ilat, lat in enumerate(lats):
            # longitude counts from high to low in Diehl grid
            for ilon, lon in reversed(list(enumerate(lons))):
                vel = grid[value][index]
                values[ilon, ilat, idepth] = vel if max_value > abs(vel) > min_value else fill_value
                index += 1
    grid = (lons, lats, depths)
    rginter = RegularGridInterpolator(grid, values, method='linear', bounds_error=False, fill_value=fill_value)
    print('Done!')
    return rginter


def get_rginter_diehl(grid, value='vps', max_value=np.inf, min_value=-np.inf, fill_value=0):
    latsU = np.unique(grid['lats'])
    lonsU = np.unique(grid['lons'])
    depthsU = np.unique(grid['depths'])

    lats = grid['lats']
    lons = grid['lons']
    depths = grid['depths']
    values = np.array(grid[value])

    indices_invalid = np.where((values < min_value) & (values > max_value))
    print('Will set {} of {} points to fill value because of min/max value.'.format(len(indices_invalid[0]), len(values)))
    values[indices_invalid] = fill_value

    # get indices to sort array with lon, lat, depth in that order
    a = np.array([lons, lats, depths])
    ind = np.lexsort((a[2, :], a[1, :], a[0, :]))

    # use indices to sort actual values and shape correspondingly
    values = values[ind]
    shape = (len(lonsU), len(latsU), len(depthsU))
    values = values.reshape(shape)

    grid = (lonsU, latsU, depthsU)
    rginter = RegularGridInterpolator(grid, values, method='linear', bounds_error=False, fill_value=fill_value)
    return rginter


def rginter_test_plot(lons, lats, depths, rginter):
    import matplotlib.pyplot as plt
    LONS, LATS = np.meshgrid(lons, lats)
    if not type(rginter) == list:
        rginter = [rginter]
    for rgi in rginter:
        fig = plt.figure()
        pc = plt.pcolormesh(LONS, LATS, rgi((LONS, LATS, depths)))
        plt.colorbar(pc)
    plt.show()


def depth_slice_test_plot(lons, lats, depths, depth, values, cmap, vmin, vmax):
    import matplotlib.pyplot as plt
    LONS, LATS = np.meshgrid(np.unique(lons), np.unique(lats))
    shape = LONS.shape
    indices = np.where(depths == depth)
    if len(indices[0]) == 0:
        raise Exception('Invalid depth')
    if not type(values) == list:
        values = [values]
    for val in values:
        fig = plt.figure()
        pc = plt.pcolormesh(LONS, LATS, val[indices].reshape(shape[:2]), cmap=cmap, vmin=vmin, vmax=vmax)
        plt.colorbar(pc)
    plt.show()


def interpolate_diehl_rde_regular_on_vgrid(vgrid_file, external_grid_file, fname_out, fname_out_vtk=None, min_depth=-4,
                                           default_cov=1.):
    ext_grid = load_json_file(external_grid_file)

    # read FMTOMO velocity/inversion grid file
    vgrid, npts, delta, start = read_vgrid(vgrid_file)

    rginter_diehl = get_rginter_diehl(ext_grid, value='rdes', fill_value=0.)

    vgrid['covs'] = np.ones(len(vgrid['vps'])) * default_cov

    tstart = datetime.now()
    print('Starting interpolation at {}'.format(tstart))
    rdes = rginter_diehl((vgrid['lons'], vgrid['lats'], vgrid['depths']))

    # set covariance ~above surface to zero!
    indices_air = np.where(np.array(vgrid['depths']) < min_depth)
    covs = transfer_rde_cov(rdes, 500.)
    covs[indices_air] = 0.1

    vgrid['covs'] = list(covs)
    vgrid['rdes'] = list(rdes)
    print('Finished after {}'.format(datetime.now() - tstart))

    write_vgrid(vgrid, npts, delta, start, fname=fname_out)
    if fname_out_vtk:
        write_vtk(vgrid, fname_out_vtk, write_data=['vps', 'covs', 'rdes'], legacy=True)


def interpolate_diehl_rde_regular_on_diehl_grid(diehl_grid_file, external_grid_file, fname_out, fname_out_vtk=None):
    ext_grid = load_json_file(external_grid_file)
    grid = load_json_file(diehl_grid_file)

    rginter_diehl = get_rginter_diehl(ext_grid, value='rdes', fill_value=0.)

    tstart = datetime.now()
    print('Starting interpolation at {}'.format(tstart))
    rdes = rginter_diehl((grid['lons'], grid['lats'], grid['depths']))
    grid['covs'] = list(transfer_rde_cov(rdes, 500.))
    grid['rdes'] = list(rdes)
    print('Finished after {}'.format(datetime.now() - tstart))

    save_json_file(grid, fname_out)
    if fname_out_vtk:
        write_vtk(grid, fname_out_vtk, write_data=['vps', 'covs', 'rdes'], legacy=True)


def interpolate_diehl_vps_regular_on_vgrid(vgrid_file, external_grid_file, fname_out, fname_out_vtk=None,
                                           default_cov=1.):
    ext_grid = load_json_file(external_grid_file)

    # read FMTOMO velocity/inversion grid file
    vgrid, npts, delta, start = read_vgrid(vgrid_file)

    rginter_diehl = get_rginter_diehl(ext_grid, value='vps', fill_value=0.)

    vgrid['covs'] = list(np.ones(len(vgrid['vps'])) * default_cov)

    tstart = datetime.now()
    print('Starting interpolation at {}'.format(tstart))
    vps = rginter_diehl((vgrid['lons'], vgrid['lats'], vgrid['depths']))
    vgrid['vps'] = list(np.array(vgrid['vps']) * (1 + vps / 100))
    print('Finished after {}'.format(datetime.now() - tstart))

    if fname_out_vtk:
        write_vtk(vgrid, fname_out_vtk, write_data=['vps'], legacy=True)
    write_vgrid(vgrid, npts, delta, start, fname=fname_out)


def interpolate_vps_regular_on_vgrid(vgrid_file, ext_grid_file, fname_out, fname_out_vtk=None, modify_vp=True,
                                     default_cov=15., ftype='npy', rel_perc=True, covs_perc=True, scale_tf=500,
                                     tf_linear=False, ext_rdes=True, min_depth=None, max_depth=None):

    print('Applying default covariance of {} percent.'.format(default_cov))

    # read FMTOMO velocity/inversion grid file
    vgrid, npts, delta, start = read_vgrid(vgrid_file)

    rginter_vp = get_rginter_tesauro_discrete(ext_grid_file, ftype=ftype, relative_values=True, col_index=3,
                                              min_depth=min_depth, max_depth=max_depth)

    #rginter_std = get_rginter_tesauro_discrete_std(ext_grid_file)

    if covs_perc:
        covs_array = np.array(vgrid['vps']) * default_cov / 100.
    else:
        covs_array = np.ones(len(np.array(vgrid['vps']))) * default_cov

    tstart = datetime.now()
    print('Starting interpolation at {}'.format(tstart))
    vps_external = rginter_vp((vgrid['lons'], vgrid['lats'], vgrid['depths']))
    vps_array = np.array(vgrid['vps'])
    ind_no_nan = ~np.isnan(vps_external)

    # interpolate covariances
    if ext_rdes:
        rginter_rde = get_rginter_tesauro_discrete(ext_grid_file, ftype=ftype, relative_values=True, col_index=4,
                                                   fill_value=0., min_depth=min_depth, max_depth=max_depth)
        rde_external = rginter_rde((vgrid['lons'], vgrid['lats'], vgrid['depths']))
        ind_no_zero_rde = rde_external > 1e-3
        covs_external = np.copy(covs_array)
        covs_external[ind_no_zero_rde] = transfer_rde_cov(rde_external, scale_tf, linear=tf_linear)[ind_no_zero_rde]
        inds_external = covs_array > covs_external
        covs_array[inds_external] = covs_external[inds_external]
        vgrid['rdes'] = list(rde_external)

    if modify_vp:
        if rel_perc:
            vps_array[ind_no_nan] = vps_array[ind_no_nan] * (1 + vps_external[ind_no_nan] / 100.)
        else:
            vps_array[ind_no_nan] = vps_external[ind_no_nan]
    else:
        print('WARNING: MODIFY VPS IS OFF.')
    vgrid['vps'] = list(vps_array)
    vgrid['covs'] = list(covs_array)

    print('Finished after {}'.format(datetime.now() - tstart))

    if fname_out_vtk:
        write_vtk(vgrid, fname_out_vtk, write_data=['vps', 'covs', 'rdes'], legacy=True)
    write_vgrid(vgrid, npts, delta, start, fname=fname_out)


def merge_tesauro_and_diehl_on_extended_diehl_grid(fn_tes, fn_diehl, fnout, _tes_only=False):
    '''
    Create an extended Diehl grid (finer than Tesauro) to interpolate Tesauro on. Then vp perturbations can be
    superimposed and smoothed before putting on coarser FMTOMO grid.
    :param fn_tes:
    :param fn_diehl: diehl differences (percentage) to ak135d
    :return:
    '''

    # def extend_axis(axis, vmin, vmax):
    #     axisU = np.unique(axis)
    #     delta = axisU[1] - axisU[0]
    #     smin = min(axisU) - vmin
    #     smax = vmax - max(axisU)
    #     nmin = int(np.ceil(smin / delta))
    #     nmax = int(np.ceil(smax / delta))
    #     vmin_new = min(axisU) - delta * nmin
    #     vmax_new = max(axisU) + delta * nmax
    #     axis_down = np.arange(vmin_new, min(axisU) - delta, delta)
    #     axis_up = np.arange(max(axisU) + delta, vmax_new, delta)
    #     axisU_new = np.append(axis_down, axisU)
    #     axisU_new = np.append(axisU_new, axis_up)
    #     return axisU_new

    # load tesauro grid and diehl grid
    lonsT, latsT, depthsT, vpsT, rdesT = np.load(fn_tes)
    lonsD, latsD, depthsD, vpsD, rdesD = np.load(fn_diehl)

    grid_diehl = grid_from_latlon_values(lonsD, latsD, depthsD, values_dict={'vps': vpsD, 'rdes': rdesD})
    rginter_tes_vps = get_rginter_tesauro_discrete(fn_tes, fill_value=0., relative_values=True, ftype='npy',
                                                   col_index=3)
    rginter_tes_rde = get_rginter_tesauro_discrete(fn_tes, fill_value=0., relative_values=True, ftype='npy',
                                                   col_index=4)
    #rginter_tes_rde = get_rginter_tesauro_discrete_rde(fn_tes, ftype='txt')
    rginter_diehl_vps = get_rginter_diehl(grid_diehl, 'vps')
    rginter_diehl_rdes = get_rginter_diehl(grid_diehl, 'rdes', fill_value=0.)

    lonMin = min(lonsT)
    lonMax = max(lonsT)
    latMin = min(latsT)
    latMax = max(latsT)
    depthsMin = min([min(depthsD), min(depthsT)])
    depthsMax = max([max(depthsD), max(depthsT)])
    delta_lonD = np.unique(lonsD)[1] - np.unique(lonsD)[0]
    delta_latD = np.unique(latsD)[1] - np.unique(latsD)[0]
    delta_depthD = np.unique(depthsD)[1] - np.unique(depthsD)[0]

    lonsU_new = np.arange(lonMin, lonMax, delta_lonD)
    latsU_new = np.arange(latMin, latMax, delta_latD)
    depthsU_new = np.arange(depthsMin, depthsMax, delta_depthD)

    lons_new, lats_new, depths_new = np.meshgrid(lonsU_new, latsU_new, depthsU_new)
    shape = lons_new.shape

    vps_tes_new = rginter_tes_vps((lons_new, lats_new, depths_new))
    rde_tes_new = rginter_tes_rde((lons_new, lats_new, depths_new))
    vps_diehl_new = rginter_diehl_vps((lons_new, lats_new, depths_new))
    rdes_diehl_new = rginter_diehl_rdes((lons_new, lats_new, depths_new))
    #cov_diehl_new = transfer_rde_cov(rdes_diehl_new, 500.)
    vps_average = weighted_average_val(vps_diehl_new, val_other=vps_tes_new, rdes=rdes_diehl_new, scale_transfer_func=1000.)
    rde_average = weighted_average_val(rdes_diehl_new, val_other=rde_tes_new, rdes=rdes_diehl_new, scale_transfer_func=1000.)
    if _tes_only:
        print('WARNING WILL ONLY USE TESAURO MESH')
        vps_average = vps_tes_new
        rde_average = rde_tes_new
    #test_plot_rginter(np.unique(lons_new), np.unique(lats_new), 7., rginter_diehl_rdes)
    #test_plot_depth_slice(lons_new, lats_new, depths_new, 49., [vps_diehl_new, vps_tes_new, vps_average],
    #                      cmap='seismic_r', vmin=-20, vmax=20)
    result = np.array([lons_new.ravel(), lats_new.ravel(), depths_new.ravel(), vps_average.ravel(), rde_average.ravel()])
    #rdes_diehl_new.ravel(), rde_tes_new.ravel()])
    #ind = np.lexsort((result[0, :], result[1, :], result[2, ::-1]))
    np.save(fnout, result)


def weighted_average_val(val_diehl, val_other, rdes, scale_transfer_func):
    weights = 1. - transfer_rde_cov(rdes, scale_transfer_func)
    val_average = weights * val_diehl + (1. - weights) * val_other
    return val_average


def grid_from_latlon_values(lons, lats, depths, values_dict):
    grid = init_dict()
    grid['lons'] = lons
    grid['lats'] = lats
    grid['depths'] = depths
    for key, value in values_dict.items():
        grid[key] = value

    return grid


def transfer_rde_cov(rdes, scale, linear=False):
    '''small transfer function based on 1/(sx**2 + 1) function'''
    if linear:
        return -1 * rdes + 1
    return 1. / (scale * rdes ** 2 + 1)


def plot_lat_slice_npy_grid(fn_grid, lat, column=3, R=6371.):
    array = np.load(fn_grid)
    lons = np.unique(array[0,:])
    lats = np.unique(array[1,:])
    depths = np.unique(array[2,:])
    # find closest lat value
    lat = lats[np.argmin(abs(lats - lat))]
    inds = np.where(array[1, :] == lat)
    values = array[column, :][inds]
    shape = (len(lons), len(depths))
    lons_km = np.deg2rad(array[0, :][inds] * R * np.cos(np.deg2rad(lat)))  # reduce radius with lat
    fig = plt.figure()
    pc = plt.pcolormesh(lons_km.reshape(shape), array[2,:][inds].reshape(shape), values.reshape(shape))
    cbar = plt.colorbar(pc)
    plt.xlabel('Km on lat = {} degree'.format(lat))
    plt.ylabel('Depth in km.')
    ax = plt.gca()
    ax.invert_yaxis()
    ax.set_aspect('equal')
    plt.show()


def plot_lon_slice_npy_grid(fn_grid, lon, column=3, R=6371.):
    array = np.load(fn_grid)
    lons = np.unique(array[0,:])
    lats = np.unique(array[1,:])
    depths = np.unique(array[2,:])
    # find closest lat value
    lon = lons[np.argmin(abs(lons - lon))]
    inds = np.where(array[0, :] == lon)
    values = array[column, :][inds]
    shape = (len(lats), len(depths))
    lats_km = np.deg2rad(array[1, :][inds] * R)
    fig = plt.figure()
    pc = plt.pcolormesh(lats_km.reshape(shape), array[2, :][inds].reshape(shape), values.reshape(shape))
    cbar = plt.colorbar(pc)
    plt.xlabel('Km on lon = {} degree'.format(lon))
    plt.ylabel('Depth in km.')
    ax = plt.gca()
    ax.invert_yaxis()
    ax.set_aspect('equal')
    plt.show()


def write_absolute(fnin_grid, fnin_diehl1d, fnout):
    with open(fnin_diehl1d, 'r') as infile:
        diehl1d = json.load(infile)
    depths = {depth: vp for depth, vp in zip(diehl1d['depths'], diehl1d['vps'])}
    grid_rel = np.load(fnin_grid)
    vprefs = np.zeros(grid_rel[0].shape)
    for depth, vpref in depths.items():
        vprefs[grid_rel[2]==depth] = vpref
    vpabs = vprefs * (1 + grid_rel[3] / 100.)
    # now it is actually grid abs!
    grid_rel[3] = vpabs
    np.save(fnout, grid_rel)


if __name__ == "__main__":
    #apply_convolution_filter('/home/marcel/marcel_scratch/alparray/diehl_2009_grid_differences_ak135.json',
    #                         wdir='/home/marcel/marcel_scratch/alparray/',
    #                         fname_out='diehl_2009_diff_filt_25_25_5.json')
    #apply_convolution_filter(12.5, 12.5, 5., (2., 2., 2.),
    #                         '/data/AlpArray_Data/alparray_grids_and_crust/diehl_2009_grid_differences_ak135.json',
    #                         wdir='/data/AlpArray_Data/alparray_grids_and_crust',
    #                         fname_out='diehl_2009_diff_filt_12.5_12.5_5.json')


    # TESAURO MODEL
    #grid = read_tesauro_model('/rscratch/minos22/marcel/alparray/grl24037-sup-0002-ds01.txt',
    #                           min_depth=-30., max_depth=70., npts_depth=21, R=6371.)
    #save_json_file(grid, '/rscratch/minos22/marcel/alparray/tesauro_discrete_5km.json')
    #write_vtk(grid, '/home/marcel/marcel_scratch/alparray/tesauro.vtk', write_cov=True, legacy=True)

    # DIEHL MODEL
    wdir='/home/marcel/marcel_scratch/alparray/'
    fn_diehl='diehl_2009_grid.json'
    fn_diehl_rel='diehl_2009_grid_differences_ak135_PERCENTAGE.json'
    fn_diehl_rel_zeros='diehl_2009_grid_diff_perc_ak135d_appended_zeros.json'
    fn_diehl_rel_filt='diehl_2009_diff_filt_8_8_6.json'
    fn_diehl_abs_filt='diehl_2009_abs_filt_8_8_6.json'
    fn_diehl_1d = '/data/AlpArray_Data/alparray_grids_and_crust/DIEHL_CRUST_GRIDS/minimum_1D_diehl.json'

    diehl_on_vgrid = False
    dts_on_vgrid = False
    kstl_on_vgrid = False
    filter_kstl_model = False
    tesauro_on_vgrid = True
    diehl_RDE_on_vgrid = False
    diehl_on_diehl = False
    abs_from_rel = False
    synthetic_box_test = False
    filter_dts_model = False
    filter_wolle_model = False
    wolle_on_vgrid = False

    #grid_differences(fname_in=fn_diehl, fname_out=fn_diehl_rel, wdir=wdir,
    #                 model=pjoin(wdir, 'minimum_1D_diehl.json'), percentage=True)
    #add_cushion_zeros_top_bot(fname_in=pjoin(wdir, fn_diehl_rel), fname_out=pjoin(wdir, fn_diehl_rel_zeros),
    #                          nlayers=10, bot=True, top=False)

    #apply_convolution_filter(8., 8., 6., spacing_llr=(2., 2., 2.),
    #                         infile=pjoin(wdir, fn_diehl_rel_zeros), wdir=wdir,
    #                         fname_out=fn_diehl_rel_filt)

    # DTS MODEL
    #fn_tesauro_rel = '/data/AlpArray_Data/alparray_grids_and_crust/DTS_CRUST_GRIDS/tesauro_SPADA_grid_dd2.0_rel_ak135dv2_PERCENTAGE_rde.npy'
    fn_tesauro_rel = '/data/AlpArray_Data/alparray_grids_and_crust/tesauro_grid_dd2.0_low_cov_ADA_rel_perc.npy'
    fn_tesauro_rel_on_dgrid = '/data/AlpArray_Data/alparray_grids_and_crust/DTS_CRUST_GRIDS/tesauro_grid_dd2.0_rel_ak135dv2_PERCENTAGE_ADA_ON_DGRID.npy'
    fn_filt_tesauro_npy = '/data/AlpArray_Data/alparray_grids_and_crust/DTS_CRUST_GRIDS/tesauro_grid_dd2.0_rel_ak135dv2_PERCENTAGE_ADA_filt_12.5_12.5_7.5.npy'
    fn_diehl_rel_rde_npy = '/data/AlpArray_Data/alparray_grids_and_crust/DIEHL_CRUST_GRIDS/diehl_2009_grid_differences_with_rde_ak135_PERCENTAGE.npy'
    #fn_diehl_rel_rde_npy = '/data/AlpArray_Data/alparray_grids_and_crust/DIEHL_CRUST_GRIDS/diehl_2009_grid_differences_ak135_with_rde_PERCENTAGE_CRUST_ONLY.npy'
    fn_diehl_tesauro_npy = '/data/AlpArray_Data/alparray_grids_and_crust/DTS_CRUST_GRIDS/diehl_and_tesauro_grid_differences_ak135_PERCENTAGE_rde.npy'
    fn_filt_diehl_tesauro_npy = '/data/AlpArray_Data/alparray_grids_and_crust/DTS_CRUST_GRIDS/diehl_and_tesauro_grid_differences_ak135_PERCENTAGE_rde_filt_12.5_12.5_7.5.npy'
    #fn_filt_diehl_tesauro_ABS_npy = '/data/AlpArray_Data/alparray_grids_and_crust/DTS_CRUST_GRIDS/diehl_and_tesauro_grid_CRUST_ONLY_rde_filt_7.5_7.5_5.npy'

    # Kaestle Model
    fn_kstl_npy = '/data/AlpArray_Data/alparray_grids_and_crust/KAESTLE_CRUST_GRIDS/kaestle_2018_PERCENTAGE_kstl_rde_100km_REGULAR_GRID_d1km.npy'
    fn_filt_kstl_npy = '/data/AlpArray_Data/alparray_grids_and_crust/KAESTLE_CRUST_GRIDS/kaestle_2018_PERCENTAGE_kstl_rde_100km_filt_12.5_12.5_7.5.npy'

    # Wolle Model
    it = 6
    #fn_wolle_npy = f'/data/AlpArray_Data/alparray_grids_and_crust/FWI_tests_Wolfgang/mk05_it{it}_rel_ak135dv2.npy'
    fn_filt_wolle_npy = None#f'/data/AlpArray_Data/alparray_grids_and_crust/FWI_tests_Wolfgang/mk05_it{it}_rel_ak135dv2_filt_11_11_10.npy'

    fn_wolle_npy = f'/data/AlpArray_Data/alparray_grids_and_crust/FWI_tests_Wolfgang/mk06_it3_rel_ak135dv2.npy'

    #merge_tesauro_and_diehl_on_extended_diehl_grid(fn_tesauro_rel, fn_diehl_rel_rde_npy, fn_diehl_tesauro_npy)

    if filter_dts_model:
        apply_convolution_filter(25., 25., 15., (2., 2., 2.), fn_diehl_tesauro_npy,
                                wdir='/data/AlpArray_Data/alparray_grids_and_crust',
                                fname_out=fn_filt_diehl_tesauro_npy,
                                values=['vps', 'rdes'], io_json=False)

    if filter_wolle_model:
        # spacing old (mk01):  (11., 11., 10.)
        apply_convolution_filter(11, 11, 10, (22.2, 15.7, 20), fn_wolle_npy,
                                wdir='/data/AlpArray_Data/alparray_grids_and_crust/FWI_tests_Wolfgang',
                                fname_out=fn_filt_wolle_npy,
                                values=['vps'], io_json=False)

    if filter_kstl_model:
        apply_convolution_filter(12.5, 12.5, 7.5, (11., 11., 1.), fn_kstl_npy,
                                wdir='/data/AlpArray_Data/alparray_grids_and_crust/KAESTLE_CRUST_GRIDS',
                                fname_out=fn_filt_kstl_npy,
                                values=['vps'], io_json=False)

    # apply_convolution_filter(7.5, 7.5, 5., (2., 2., 2.), fn_diehl_tesauro_npy,
    #                         wdir='/data/AlpArray_Data/alparray_grids_and_crust',
    #                         fname_out=fn_filt_diehl_tesauro_npy,
    #                         values=['vps', 'rdes'], io_json=False)
    # write_absolute(fn_filt_diehl_tesauro_npy, fn_diehl_1d, fn_filt_diehl_tesauro_ABS_npy)

    if dts_on_vgrid:
        #fn_tesauro_discrete = '/data/AlpArray_Data/alparray_grids_and_crust/tesauro_SPADA_grid_dd2.0.txt'
        #fpath = '/data/AlpArray_Data/fmtomo/v6/crust_incl_lf_sm_FIX_DTS_COARSE_grad_sm30_dm10'
        fpath = '/data/AlpArray_Data/fmtomo/v6_S/crust_incl_hf_sm_FIX_DTS_grad_sm6_dm10'
        #fpath = '/data/AlpArray_Data/fmtomo/v5_resolution_analysis/crust_incl_hf_sm_FIX_DTS_grad_sm30_dm10_checkerboard_3x3x4'
        interpolate_vps_regular_on_vgrid(pjoin(fpath, 'vgridsref_orig.in'), fn_filt_diehl_tesauro_npy,
                                         pjoin(fpath, 'vgrids_dts_crust.in'),
                                         fname_out_vtk=pjoin(fpath, 'vgrids_dts_crust.vtk'),
                                         default_cov=15., ftype='npy', covs_perc=True, modify_vp=True)

    if kstl_on_vgrid:
        fpath = '/data/AlpArray_Data/fmtomo/v6_S/crust_incl_hf_KSTL_grad_sm6_dm10_S_SKS'
        #fpath = '/data/AlpArray_Data/fmtomo/v6_S/no_crust_correction_hf_grad_sm6_dm10'
        interpolate_vps_regular_on_vgrid(pjoin(fpath, 'vgridsref_orig.in'), fn_filt_kstl_npy,
                                         pjoin(fpath, 'vgrids_kstl_crust.in'),
                                         fname_out_vtk=pjoin(fpath, 'vgrids_kstl_crust.vtk'),
                                         default_cov=15., ftype='npy', covs_perc=True, modify_vp=True, ext_rdes=True,
                                         tf_linear=True)

    if wolle_on_vgrid:
        fpath = '/data/AlpArray_Data/fmtomo/v6/fwi_model_wolfgang_test'
        grid_file = fn_filt_wolle_npy if filter_wolle_model else fn_wolle_npy
        interpolate_vps_regular_on_vgrid(vgrid_file=pjoin(fpath, 'vgridsref_orig.in'), ext_grid_file=grid_file,
                                         fname_out=pjoin(fpath, f'vgrids_wolle_mk06_3.in'),
                                         fname_out_vtk=pjoin(fpath, f'vgrids_wolle_mk06_3.vtk'),
                                         default_cov=15., ftype='npy', covs_perc=True, modify_vp=True, ext_rdes=False,
                                         tf_linear=True, max_depth=None)


    if tesauro_on_vgrid:
        # filter Tesauro/Spada only !!!
        merge_tesauro_and_diehl_on_extended_diehl_grid(fn_tesauro_rel, fn_diehl_rel_rde_npy, fn_tesauro_rel_on_dgrid,
                                                       _tes_only=True)
        apply_convolution_filter(12.5, 12.5, 7.5, (2., 2., 2.), fn_tesauro_rel_on_dgrid,
                                 wdir='/data/AlpArray_Data/alparray_grids_and_crust',
                                 fname_out=fn_filt_tesauro_npy,
                                 values=['vps', 'rdes'], io_json=False)

        fpath = '/data/AdriaArray_Data/fmtomo_adriaarray/alpadege/crust_incl_TESAURO_sm30_dm1'
        interpolate_vps_regular_on_vgrid(pjoin(fpath, 'vgrids.in'), fn_filt_tesauro_npy,
                                         pjoin(fpath, 'vgrids_tesauro_crust.in'),
                                         fname_out_vtk=pjoin(fpath, 'vgrids_tesauro_crust.vtk'),
                                         default_cov=15., ftype='npy', covs_perc=True)

    fpath = '/rscratch/minos13/marcel/fmtomo_alparray/v4/different_test_runs/alparray_mantle_diehl_crust_included_hf_1D_model_altered'

    if diehl_on_vgrid:
        interpolate_diehl_vps_regular_on_vgrid(pjoin(fpath, 'vgridsref_no_crust.in'), pjoin(wdir, fn_diehl_rel_filt),
                                               fname_out=pjoin(fpath, 'vgrids_crust.in'),
                                               fname_out_vtk=pjoin(fpath, 'vgrids_diehl_crust.vtk'))

    # read diehl layers with RDE values and interpolate from regular grid on grid
    external_grid_file = '/data/AlpArray_Data/crust_diehl_2009/layers_filtered_12.5_12.5_0.1-2_2_15.json'
    #vgrid_file = '/rscratch/minos13/marcel/fmtomo_alparray/v3.5/alparray_mantle_diehl_crust_included_hf_gradient_smoothing/vgridsref_crust.in'

    fname_in = 'vgrids_crust.in' # IF NOT DOING CRUSTAL CORRECTIONS CHANGE THIS TO vgrids with crust: vgrids_crust.in
    fname_out =  pjoin(fpath, 'vgridsref_diehl_crust_cov.in')
    fname_out_vtk = pjoin(fpath, 'vgridsref_diehl_crust_cov.vtk')
    #grid = read_diehl_rde_layers()
    #grid_conv = convolution_filter(grid, (12.5, 12.5, 0.1), spacing_llr=(2, 2, 15),
    #                               wdir='/home/marcel/marcel_scratch/alparray', val='rdes')
    #save_json_file(grid, external_grid_file)

    if abs_from_rel:
        # get absolute velocities on crustal model from smoothed relative values
        apply_vel_changes_to_absolute_grid(pjoin(wdir, fn_diehl), pjoin(wdir, fn_diehl_rel_filt),
                                           fname_out=pjoin(wdir, fn_diehl_abs_filt))

    if diehl_RDE_on_vgrid:
        # a priori covariance on vgrid based on RDE valeus
        interpolate_diehl_rde_regular_on_vgrid(pjoin(fpath, fname_in), external_grid_file, fname_out,
                                               fname_out_vtk=pjoin(fpath, fname_out_vtk))

    if diehl_on_diehl:
        # a priori covariances on crust grid (e.g. for plotting threshold)
        interpolate_diehl_rde_regular_on_diehl_grid(pjoin(wdir, fn_diehl_rel_filt), external_grid_file,
                                                    pjoin(wdir, 'diehl_2009_grid_diff_with_rdes_12.5_12.5_0.1.json'),
                                                    fname_out_vtk=pjoin(wdir, 'diehl_2009_rdes.vtk'))

    # Synthetic Box test (grid spacing for boxes ~10x10x10 km)
    # discretize_box('/home/marcel/marcel_scratch/alparray/box2.1_legacy.vtk',
    #                '/home/marcel/marcel_scratch/alparray/box2.1.json',
    #                vp_diff=3., R=6371., dr=10., dLat=0.09, dLon=0.09)

    fpath = '/data/AlpArray_Data/fmtomo/v5_resolution_analysis/crust_incl_hf_sm_FIX_DTS_grad_sm30_dm10_profile_2e_EASI_detached_slab_N'
    wdir = '/data/AlpArray_Data/sciebo/AlpArray_home/resolution_analysis/slice_2e_delamination'


    if synthetic_box_test:
        fn_vgrid_out = pjoin(fpath, 'vgridsref_crust_slab_E_dip_N.in')
        for index in range(2):
            if index == 0:
                fn_box_vtk = pjoin(wdir, 'box_vminus_delamin.vtk')
                fn_box_json = pjoin(wdir, 'box_vminus_delamin.json')
                fn_box_filt = pjoin(wdir, 'box_vminus_delamin_12.5_12.5_7.5.json')
                vp_diff = -4.
                fn_vgrid = pjoin(fpath, 'vgrids_dts_crust_cov_VGREF.in')
            elif index == 1:
                fn_box_vtk = pjoin(wdir, 'box_slab_E_dip_N_delamin.vtk')
                fn_box_json = pjoin(wdir, 'box_slab_E_dip_N_delamin.json')
                fn_box_filt = pjoin(wdir, 'box_slab_E_dip_N_12.5_12.5_7.5_delamin.json')
                vp_diff = 6.
                fn_vgrid = fn_vgrid_out

            # interpolate box on a regular grid
            discretize_box(fn_box_vtk, fn_box_json, vp_diff=vp_diff, R=6371., dr=5., dLat=0.045, dLon=0.045)

            # smooth this grid
            apply_convolution_filter(12.5, 12.5, 7.5, (5., 5., 5.), fn_box_json, wdir=wdir, fname_out=fn_box_filt,
                                     io_json=True)

            interpolate_on_vgrid(fn_vgrid, fn_box_filt,
                                 fname_out=fn_vgrid_out, vel_perturbation_abs=True, ncores=20)

    # read and overwrite fn_vgrid_out with covariances from RDE values
    # ... not necessary because forward grid for resolution analysis is not reference grid with covariances
    #interpolate_diehl_rde_regular_on_vgrid(fn_vgrid_out, external_grid_file, fn_vgrid_out,
    #                                       fname_out_vtk=pjoin(fpath, 'vgridsref_crust_box_lip_b_south.vtk'))

    # apply_convolution_filter(12.5, 12.5, 10., (10., 10., 10.),
    #                          '/home/marcel/marcel_scratch/alparray/box2.2.json',
    #                          wdir='/home/marcel/marcel_scratch/alparray/',
    #                          fname_out=pjoin'box2.2_grid_12.5_12.5_10.json')
    # apply_convolution_filter(12.5, 12.5, 10., (10., 10., 10.),
    #                          '/home/marcel/marcel_scratch/alparray/box2.3.json',
    #                          wdir='/home/marcel/marcel_scratch/alparray/',
    #                          fname_out=pjoin'box2.3_grid_12.5_12.5_10.json')

    #apply_convolution_filter(12.5, 12.5, 10., (10., 10., 10.),
    #                         '/home/marcel/marcel_scratch/alparray/box3.3.json',
    #                         wdir='/home/marcel/marcel_scratch/alparray/',
    #                         fname_out=pjoin'box3.3_grid_12.5_12.5_10.json')

    #interpolate_on_vgrid('/rscratch/minos13/marcel/fmtomo_alparray/alparray_mantle_from_m6.0_diehl_crustal_corrections_v2_template/vgridsref.in',
    #                     '/rscratch/minos22/marcel/alparray/box2.1_grid_12.5_12.5_10.json',
    #                     fname_out='/rscratch/minos22/marcel/alparray/vgrids_box2.1.in',
    #                     vel_perturbation_abs=True, ncores=20)
    # interpolate_on_vgrid('/rscratch/minos22/marcel/alparray/vgrids_box2.1.in',
    #                      '/rscratch/minos22/marcel/alparray/box2.2_grid_12.5_12.5_10.json',
    #                      fname_out='/rscratch/minos22/marcel/alparray/vgrids_box2.2.in',
    #                      vel_perturbation_abs=True, ncores=20)
    # interpolate_on_vgrid('/rscratch/minos22/marcel/alparray/vgrids_box2.2.in',
    #                      '/rscratch/minos22/marcel/alparray/box2.3_grid_12.5_12.5_10.json',
    #                      fname_out='/rscratch/minos22/marcel/alparray/vgrids_three_boxes.in',
    #                      fname_out_vtk='/rscratch/minos22/marcel/alparray/vgrids_three_boxes.vtk',
    #                      vel_perturbation_abs=True, ncores=20)
    #interpolate_on_vgrid('/rscratch/minos22/marcel/alparray/vgrids_box2.2.in',
    #                     '/rscratch/minos22/marcel/alparray/box3.3_grid_12.5_12.5_10.json',
    #                     fname_out='/rscratch/minos22/marcel/alparray/vgrids_three_boxes_inverted.in',
    #                     fname_out_vtk='/rscratch/minos22/marcel/alparray/vgrids_three_boxes_inverted.vtk',
     #                    vel_perturbation_abs=True, ncores=20)


    #interpolate_on_vgrid(
    #    '/rscratch/minos13/marcel/fmtomo_alparray/alparray_mantle_from_m6.0_diehl_crustal_corrections_v2_sm10_damp10/it_6/vgrids.in',
    #    '/rscratch/minos22/marcel/alparray/diehl_2009_diff_filt_12.5_12.5_10.json',
    #    fname_out='vgrids_it6_with_diehl.in',
    #    fname_out_vtk=None,
    #    vel_perturbation_abs=True, ncores=20,
    #    wdir='/rscratch/minos13/marcel/fmtomo_alparray/alparray_mantle_from_m6.0_diehl_crustal_corrections_v2_sm10_damp10/')


    #tomo_sim_dir = '/home/marcel/marcel_scratch/asp3d_elvis/simu02'
    #tomo_sim_dir = '/home/marcel/marcel_scratch/alparray/fmtomo_traveltime_tomo/example3'
    #tomo_sim_dir = r'/rscratch/minos22/marcel/alparray/fmtomo_traveltime_tomo/alparray_0/'

    #interfaces2vtk(os.path.join(tomo_sim_dir, 'interfaces.in'), outpath=tomo_sim_dir)
    #propgrid2vtk(os.path.join(tomo_sim_dir, 'propgrid.in'), os.path.join(tomo_sim_dir, 'propgrid.vtk'))
    #vgrids2VTK(r'/rscratch/minos22/marcel/alparray/fmtomo_traveltime_tomo/alparray_0/vgrids.in', 'vgrids.vtk') # different coord system and units





