#!/usr/bin/env python
# -*- coding: utf-8 -*-
import os

import numpy as np
from scipy.interpolate import RegularGridInterpolator
import matplotlib.pyplot as plt

from pylot.tomography.fmtomo_tools.fmtomo_teleseismic_utils import export_otimes, organize_receivers

def crustal_correction_using_differences(fname_otimes, fname_otimes_diff, fname_otimes_out):
    otimes = np.loadtxt(fname_otimes, skiprows=1)
    otimes_diff = np.loadtxt(fname_otimes_diff, skiprows=1)

    src_means = source_means(otimes_diff)

    #mean_diffs = np.mean([float(item.split()[4]) for item in otimes_diff[1:]])
    print('Mean_diffs of all sources: ', np.mean(list(src_means.values())))
    # TODO: Check if demeaning is useful for crustal correction using synthetics (was it b4 using ak135_diehl 1D?)
    # Update 27.5.2020: If not demeaned, overall velocity perturbation shifted to negative in whole upper mantle
    # BUT: demean has to be executed for each source, not with global mean!

    with open(fname_otimes_out, 'w') as outfile:
        outfile.write('{}\n'.format(len(otimes)))
        for index, line in enumerate(otimes):
            src_id = line[1]
            ttime = line[4]
            ttime_diff = otimes_diff[index][4]
            new_time = ttime - ttime_diff + src_means[src_id]
            line[4] = new_time
            for col_index, item in enumerate(line):
                if col_index < 4:
                    fmt = '{:10g} '
                else:
                    fmt = '{:10.8f} '
                outfile.write(fmt.format(item))
            outfile.write('\n')


def source_means(otimes):
    # catch all source ids from otimes file
    src_ids = np.unique(otimes[:, 1])
    print('Source IDs:', src_ids)

    # create dictionary containing means of travel time (differences) for each source (key)
    src_means = {src_id: np.mean(otimes[otimes[:, 1] == src_id][:, 4]) for src_id in src_ids}

    print('Source means:')
    for srcid, val in src_means.items():
        print(srcid, ': ', val)

    return src_means


def crustal_correction_using_residual_map(fname_otimes, fname_residuals, receivers_in, fname_otimes_out):
    ''' correct using residuals from numpy file with an array of shape (lons, lats, residuals)'''
    otimes = np.loadtxt(fname_otimes, skiprows=1)
    rec_dict = organize_receivers(receivers_in)
    lons, lats, res = np.load(fname_residuals, allow_pickle=True)

    lonsU = np.unique(lons)
    latsU = np.unique(lats)
    #grid = (lonsU, latsU)
    #shape = [len(item) for item in grid]
    #res = res.reshape(shape)

    # not very efficient but very lazy (just interpolate value for each ray, i.e. receiver, not unique,
    # which is quite redundant)
    rginter = RegularGridInterpolator((lonsU, latsU), res, bounds_error=False, fill_value=0.)

    test_rginter(rginter)
    
    for line in otimes:
        rec_id = int(line[0])
        lonlat = (rec_dict[rec_id]['lon'], rec_dict[rec_id]['lat'])
        res = rginter(lonlat)
        line[4] -= res

    src_means = source_means(otimes)

    for src_id, src_mean in src_means.items():
        otimes[otimes[:, 1] == src_id, 4] -= src_mean

    with open(fname_otimes_out, 'w') as outfile:
        outfile.write('{}\n'.format(len(otimes)))
        for line in otimes:
            for col_index, item in enumerate(line):
                if col_index < 4:
                    fmt = '{:10g} '
                else:
                    fmt = '{:10.8f} '
                outfile.write(fmt.format(item))
            outfile.write('\n')


def test_rginter(rginter):
    lons = np.linspace(0, 22, 220)
    lats = np.linspace(40, 52, 120)
    lonsg, latsg = np.meshgrid(lons, lats)
    data = rginter((lonsg, latsg))
    pcm = plt.pcolormesh(lonsg, latsg, data)
    plt.colorbar(pcm)
    plt.show()


def get_otimes_diff(fname_arrivals1, fname_arrivals2, fname_out='otimes_diff.dat'):
    '''
    Calculate file containing travel time differences between fname_arrivals1 and fname_arrivals2, e.g. for crustal
    correction.
    :param fname_arrivals1:
    :param fname_arrivals2:
    :return:
    '''
    with open(fname_arrivals1, 'r') as infile:
        arrivals1 = infile.readlines()
    with open(fname_arrivals2, 'r') as infile:
        arrivals2 = infile.readlines()

    assert len(arrivals1) == len(arrivals2), 'Length of input arrival files differs'

    print('Calculating differences between file {} and {}'.format(fname_arrivals1, fname_arrivals2))
    columnString = '{}    '

    nArrivals = len(arrivals1)
    with open(fname_out, 'w') as outfile:
        outfile.write('{}\n'.format(nArrivals))
        for index in range(nArrivals):
            line1 = arrivals1[index]
            line2 = arrivals2[index]
            for item in line1.split()[:4]:
                outfile.write(columnString.format(item))
            diff = float(line1.split()[4]) - float(line2.split()[4])
            outfile.write(columnString.format(diff))
            outfile.write('\n')
    print('Wrote {} lines to file {}'.format(nArrivals, fname_out))


def get_synthetic_obsdata_legacy(fname_m1, fname_m2, fname_otimes_orig, fname_out='otimes_modif.dat'):#, p_err=0.1):
    '''
    Create synthetic obsdata of model 1 relative to model 2 (usually 1D model synthetic travel times)
    :param fname_m1: arrivals.dat file of synthetic travel times (e.g. from checkerboard test)
    :param fname_m2: arrivals.dat file of synthetic travel times (e.g. ak135 model)
    :param fname_otimes_orig: original otimes file for pick uncertainties
    :param fname_out: otimes.dat file with output
    :param p_err: picking error
    :return:
    '''
    with open(fname_m1, 'r') as infile:
        arrivals_model1 = infile.readlines()
    with open(fname_m2, 'r') as infile:
        arrivals_model2 = infile.readlines()
    with open(fname_otimes_orig, 'r') as infile:
        otimes = infile.readlines()[1:]

    assert(len(arrivals_model1) == len(arrivals_model2) == len(otimes)), 'missmatch in lengths of files!'

    diffs = []
    with open(fname_out, 'w') as outfile:
        outfile.write('{}\n'.format(len(arrivals_model1)))

        for line_model1, line_model2, line_otimes in zip(arrivals_model1, arrivals_model2, otimes):
            for item in line_model2.split()[:4]:
                outfile.write('{}    '.format(item))

            ttime_model1 = float(line_model1.split()[4])
            ttime_model2 = float(line_model2.split()[4])
            pickerror = float(line_otimes.split()[5])
            # pickerror = (p_err+2*p_err*abs(np.random.randn()))

            ttime_diff = ttime_model1 - ttime_model2
            diffs.append(ttime_diff)

            outfile.write('{}    {}\n'.format(ttime_diff, pickerror))

    mean_diff = np.mean(diffs)
    abs_mean_diff = np.mean(np.abs(diffs))

    print('Mean_diff: {} - abs_mean_diff: {}'.format(mean_diff, abs_mean_diff))
    print('Done with {} travel times. Output in file: {}'.format(len(arrivals_model1), fname_out))


def get_synthetic_obsdata(fname_m1, fname_m2, fname_otimes_orig, fname_out='otimes_modif.dat', sigma='original'):
    '''
    Create synthetic obsdata of model 1 relative to model 2 (usually 1D model synthetic travel times)
    :param fname_m1: arrivals.dat file of synthetic travel times (e.g. from checkerboard test)
    :param fname_m2: arrivals.dat file of synthetic travel times (e.g. ak135 model)
    :param fname_otimes_orig: original otimes file for pick uncertainties
    :param fname_out: otimes.dat file with output
    :param p_err: picking error
    :return:
    '''
    arrivals_model1 = np.genfromtxt(fname_m1)
    arrivals_model2 = np.genfromtxt(fname_m2)
    otimes = np.genfromtxt(fname_otimes_orig, skip_header=1)

    assert(len(arrivals_model1) == len(arrivals_model2) == len(otimes)), 'missmatch in lengths of files!'

    # get new array as copy from model1 after deleting last column (not needed for otimes)
    arrivals_diff = np.delete(arrivals_model1, 6, 1)

    # overwrite time column by diffs
    arrivals_diff[:, 4] -= arrivals_model2[:, 4]
    # overwrite last column by pick errors
    arrivals_diff[:, 5] = otimes[:, 5]

    # get max nSrc (in last line of sorted arrivals files)
    nSrc = int(arrivals_diff[-1, 1])

    print('N sources:', nSrc)
    if sigma not in [None, False]:
        print('Applying gaussian noise with sigma={}'.format(sigma))
        if sigma == 'original':
            sigma_val = otimes[:, 5]
        else:
            sigma_val = sigma
        gauss_noise = np.random.normal(0., sigma_val, len(arrivals_diff[:, 4]))
        plt.axhline(0., color='grey')
        plt.vlines(np.arange(0, len(arrivals_diff)), arrivals_diff[:, 4], arrivals_diff[:, 4] + gauss_noise,
                   color='grey', linestyles='dashed')
        plt.plot(arrivals_diff[:, 4], 'r.')
        arrivals_diff[:, 4] += gauss_noise

    for index in range(nSrc):
        src_id = index + 1
        # get indices for this source ID
        indices = np.where(arrivals_diff[:, 1] == src_id)
        # get mean for that srcid
        mean = np.mean(arrivals_diff[:, 4][indices])
        print('Srcid: {}, mean: {}'.format(src_id, mean))
        # de-mean array
        arrivals_diff[:, 4][indices] -= mean

    if sigma not in [None, False]:
        plt.axhline(0., color='grey')
        #plt.vlines(np.arange(0, len(arrivals_diff)), arrivals_diff[:, 4], arrivals_diff[:, 4] + gauss_noise)
        plt.plot(arrivals_diff[:, 4], 'b.')
        plt.figure()
        plt.hist(gauss_noise, bins=100)
        plt.show()

    export_otimes(arrivals_diff, fname_out)


def plot_histograms_for_crustal_corrections():
    import matplotlib.pyplot as plt
    ttdiffs_no_crust = []
    ttdiffs_crust = []
    ttdiffs_corrected = []

    #get_synthetic_obsdata('it_synth_forward/arrivals.dat', 'arrivals_ak135_diehl.dat', 'otimes_three_boxes_no_crust.dat')
    #get_synthetic_obsdata('it_synth_forward_with_crust/arrivals.dat', 'arrivals_ak135_diehl.dat', 'otimes_three_boxes_with_diehl_crust.dat')
    with open('otimes_three_boxes_with_diehl_crust.dat', 'r') as infile:
        lines_crust = infile.readlines()[1:]
    with open('otimes_three_boxes_no_crust.dat', 'r') as infile:
        lines_no_crust = infile.readlines()[1:]
    with open('otimes_three_boxes_crustal_corrected.dat', 'r') as infile:
        lines_corrected = infile.readlines()[1:]

    for line in lines_crust:
        ttdiffs_crust.append(float(line.split()[4]))
    for line in lines_no_crust:
        ttdiffs_no_crust.append(float(line.split()[4]))
    for line in lines_corrected:
        ttdiffs_corrected.append(float(line.split()[4]))

    plt.hist(ttdiffs_crust, bins=100, label='tt-diff including crustal structure.')
    plt.hist(ttdiffs_no_crust, bins=100, label='tt-diff without crustal structure')
    plt.hist(ttdiffs_corrected, bins=100, label='tt-diff after correcting for crustal structure')
    plt.xlabel('Travel time differences to ak135_diehl 1D model [s]')
    plt.legend()

if __name__ == '__main__':
    #wdir = '/rscratch/minos13/marcel/fmtomo_alparray/alparray_mantle_diehl_crustal_corrections_v3_hf'
    #fname1 = os.path.join(wdir, 'arrivals_ak135di_crust.dat')
    #fname2 = os.path.join(wdir, 'arrivals_ak135di.dat')
    #get_otimes_diff(fname1, fname2, os.path.join(wdir, 'otimes_diff.dat'))

    #get_synthetic_obsdata('arrivals_cb_N2.dat', 'arrivals_ak135.dat', 'otimes_cb_N2.dat')
    #os.chdir('/data/AlpArray_Data/fmtomo/v4/different_test_runs/alparray_mantle_waldhauser_crust_corrected_covs_hf_gradient_smoothing')
    #os.chdir('/data/AlpArray_Data/fmtomo/v4/different_test_runs/alparray_mantle_waldhauser_crust_corrected_covs_hf_gradient_smoothing')
    #crustal_correction_using_residual_map('otimes.dat', 'wh_residuals.npy', 'receivers.in', 'otimes_corr_wh.dat')

    #os.chdir('/data/AlpArray_Data/fmtomo/v4/different_test_runs/alparray_mantle_diehl_crust_corrected_residuals_stacked_gradient_smoothing')
    #crustal_correction_using_residual_map('otimes.dat', 'diehl_2009_residuals.npy', 'receivers.in', 'otimes_corr_diehl.dat')

    #os.chdir('/data/AlpArray_Data/fmtomo/v6/crust_corrected_VERTICAL_hf_sm_FIX_DTS_grad_sm30_dm10')
    #crustal_correction_using_residual_map('otimes.dat', 'dts_filt_12.5_12.5_7.5_CRUST_ONLY_RESIDUALS.npy', 'receivers.in', 'otimes_corr_dts.dat')

    os.chdir('/data/AlpArray_Data/fmtomo/v6/crust_corrected_WH_hf_sm_FIX_DTS_grad_sm30_dm10')
    crustal_correction_using_residual_map('otimes.dat', 'wh_residuals_SORTED.npy', 'receivers.in', 'otimes_corr_wh.dat')



    #crustal_correction_using_residual_map('otimes.dat', 'wh_residuals.npy', 'receivers.in', 'otimes_corr_wh.dat')