# This script estimates density of station distribution using Gaussian KDE. For each point a Gaussian Kernel with
# sigma ~ grid size (?) is summed in 2D, calculating estimate density function. If the sum of all points evaluated
# at the data points increases that means there are more regions with high density

import os
import copy

import matplotlib.pyplot as plt
import numpy as np
from scipy.stats import gaussian_kde

from pylot.tomography.fmtomo_tools.fmtomo_teleseismic_utils import organize_receivers, organize_event_names, export_otimes

def get_events_dict(otimes, recs):
    events_dict = {}
    for otime in otimes:
        rec_id, src_id = int(otime[0]), int(otime[1])
        if not src_id in events_dict.keys():
            events_dict[src_id] = dict(lons=[], lats=[], rec_ids=[])
        events_dict[src_id]['lons'].append(recs[rec_id]['lon'])
        events_dict[src_id]['lats'].append(recs[rec_id]['lat'])
        events_dict[src_id]['rec_ids'].append(rec_id)
    return events_dict


def transfer_kde_to_weight(kdes):
    weights = 0.1 - kdes
    indices_negative = np.where(weights < 0)[0]
    if indices_negative:
        print('Warning! {} negative indices in weights. Set to 0.'.format(len(indices_negative)))
        weights[indices_negative] = 0.
    return weights


def calc_event_kernels(events_dict, eventnames, kde_factor, plot_single=False, fdir_out=None):
    kernels = {}
    evaluated_kernels = {}

    if plot_single:
        fig = plt.figure(figsize=(16, 9))

    for index in range(len(eventnames)):
        eventname = eventnames[index]
        source = index + 1
        lons = events_dict[source]['lons']
        lats = events_dict[source]['lats']
        kernel = gaussian_kde(np.array([lons, lats]), bw_method=kde_factor)
        evaluated_kernel = kernel((lons, lats))
        kernels[source] = kernel
        evaluated_kernels[source] = evaluated_kernel
        if plot_single:
            assert fdir_out, 'Need to specify output directory for plots'
            if not os.path.isdir(fdir_out):
                os.mkdir(fdir_out)
            weights = transfer_kde_to_weight(evaluated_kernel)
            iterdict = {'kde': evaluated_kernel, 'weight': weights}
            for name, colorarray in iterdict.items():
                if name == 'kde':
                    vmin=0
                    vmax=0.05
                else:
                    vmin=None
                    vmax=None
                scs = plt.scatter(lons, lats, marker='o', c=colorarray, lw=1.0, vmin=vmin, vmax=vmax)
                cbar = plt.colorbar(scs)
                plt.title('Station {} for event {}'.format(name, eventname))
                fig.savefig(os.path.join(fdir_out, '{}_{}.svg'.format(eventname, name)))
                fig.clear()
                #plt.show()
    if plot_single:
        plt.close('all')
    return kernels, evaluated_kernels


def plot_hist_kernels(evaluated_kernels):
    all_kernels = np.array([])
    for kernel in evaluated_kernels.values():
        all_kernels = np.append(all_kernels, kernel)
    plt.hist(all_kernels, bins=200, label='kernels')
    plt.title('Distribution of all kernels.')
    plt.show()


def plot_hist_uncertainties(otimes_new, otimes_orig, bins=200):
    plt.hist(otimes_orig[:, -1], lw=1, bins=bins, label='Unmodified uncertainties', fc=(.5, .5, 0., 0.5))
    plt.hist(otimes_new[:, -1], lw=1, bins=bins, label='New uncertainties', fc=(0., .5, .5, 0.5))
    plt.legend()
    plt.xlabel('Uncertainty [s]')
    plt.title('Uncertainty distribution before and after correction.')
    plt.show()


def plot_uncertainties(otimes, recs, eventnames, fdir_out, vmin=0.1, vmax=0.4):
    if not os.path.isdir(fdir_out):
        os.mkdir(fdir_out)

    print('Writing output to directory:', fdir_out)
    epsilon = 1e-6

    fig = plt.figure(figsize=(16, 9))
    for index, eventname in enumerate(eventnames):
        src_id = index + 1
        indices = np.where(abs(otimes[:, 1] - src_id) <= epsilon)
        rec_ids = otimes[:, 0][indices]
        uncs = otimes[:, -1][indices]
        lats = [recs[rec_id]['lat'] for rec_id in rec_ids]
        lons = [recs[rec_id]['lon'] for rec_id in rec_ids]
        sc = plt.scatter(lons, lats, c=uncs, vmin=vmin, vmax=vmax)
        plt.title(eventname)
        cb = plt.colorbar(sc)
        fig.savefig(os.path.join(fdir_out, '{}.svg'.format(eventname)))
        fig.clear()
    plt.close('all')


def exp_func(uncertainty, eval_kernel, exponent=30):
    return uncertainty * np.exp(exponent * eval_kernel)


def modify_otimes(otimes, recs, kernels):
    print('Applying Kernel on otimes and modifying uncertainties...')
    for otime in otimes:
        rec_id, src_id = int(otime[0]), int(otime[1])
        lon = recs[rec_id]['lon']
        lat = recs[rec_id]['lat']
        eval_kernel = kernels[src_id]((lon, lat))
        otime[-1] = exp_func(otime[-1], eval_kernel)
    return otimes


def plot_average_kernel(evaluated_kernels, eventnames):
    ids_names = [(index + 1, eventname) for index, eventname in enumerate(eventnames)]
    sorted_events = sorted(ids_names, key=lambda x: x[1])
    kernel_sums = {src_id: np.sum(kernel) for src_id, kernel in evaluated_kernels.items()}
    src_ids = [item[0] for item in sorted_events]
    eventnames = [item[1] for item in sorted_events]
    sums = [kernel_sums[src_id] for src_id in src_ids]
    plt.plot(sums)
    xticks = np.arange(0, len(eventnames), step=len(eventnames)//10)
    xticklabels = [eventname[:8] for index, eventname in enumerate(eventnames) if index in xticks]
    plt.xticks(xticks, xticklabels)
    plt.title('Average kernel per station. High value means inhomogeneous station distribution.')
    plt.show()


def export_station_kdes(events_dict, evaluated_kernels, fnout):
    with open(fnout, 'w') as outfile:
        for src_id in range(1, len(events_dict) + 1):
            kernel = evaluated_kernels[src_id]
            for kde, rec_id in zip(kernel, events_dict[src_id]['rec_ids']):
                outfile.write('{}    {}    {}\n'.format(rec_id, src_id, kde))


def main(plot=False, plot_detailed=False):
    working_path = '/rscratch/minos13/marcel/fmtomo_alparray/v4/alparray_mantle_diehl_crust_included_hf_gradient_smoothing'
    fdir_out = 'station_density'
    fnout = os.path.join(fdir_out, 'station_kdes.txt')

    os.chdir(working_path)
    eventnames = organize_event_names('input_source_file_P.in')
    recs = organize_receivers('receivers.in')
    otimes_data = np.genfromtxt('otimes_orig.dat', skip_header=1)

    # kernel width ~ grid size? or station spacing?
    kde_size = 30  # km
    kde_factor = kde_size / (6371. * np.pi / 180.)
    print('KDE width (degree):', kde_factor)

    events_dict = get_events_dict(otimes_data, recs)
    kernels, evaluated_kernels = calc_event_kernels(events_dict, eventnames, kde_factor, plot_single=plot_detailed,
                                                    fdir_out=fdir_out)

    export_station_kdes(events_dict, evaluated_kernels, fnout)

    otimes_modif = modify_otimes(copy.deepcopy(otimes_data), recs, kernels)
    #export_otimes(otimes_modif, os.path.join(working_path, 'otimes_modif_kernel.dat'))
    if plot:
        plot_average_kernel(evaluated_kernels, eventnames)
        plot_hist_kernels(evaluated_kernels)
        plot_hist_uncertainties(otimes_modif, otimes_data)
    if plot_detailed:
        plot_uncertainties(otimes_data, recs, eventnames, 'uncerts_old')
        plot_uncertainties(otimes_modif, recs, eventnames, 'uncerts_new')


if __name__ == '__main__':
    main(plot=True, plot_detailed=True)
